# -*- coding: utf-8 -*-
"""
reV-ReEDS-PLEXOS command line interface (cli).
"""
import os
import pandas as pd
import click
import logging

from rex.utilities.hpc import SLURM
from rex.utilities.cli_dtypes import STR, INT, INTLIST
from rex.utilities.loggers import init_mult

from reVX.plexos.rev_reeds_plexos import Manager

logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.option('--name', '-n', default='plx', type=STR,
              help='Job name. Default is "plx".')
@click.option('--job_input', '-j', required=True,
              type=click.Path(exists=True),
              help=('Path to plexos job input csv.'))
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--reeds_dir', '-rd', required=True, type=click.Path(),
              help='Directory containing REEDS buildout files.')
@click.option('--cf_years', '-y', required=True, type=INTLIST,
              help='Capacity factor resource year.')
@click.option('--build_years', '-by', required=True, type=INTLIST,
              help='REEDS build years to aggregate profiles for.')
@click.option('--scenario', '-s', default=None, type=STR, show_default=True,
              help='Optional filter to run just one scenario from job input.')
@click.option('-v', '--verbose', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, job_input, out_dir, reeds_dir, cf_years, build_years,
         scenario, verbose):
    """reV-ReEDS-PLEXOS Command Line Interface"""

    ctx.ensure_object(dict)
    ctx.obj['NAME'] = name
    ctx.obj['JOB_INPUT'] = job_input
    ctx.obj['OUT_DIR'] = out_dir
    ctx.obj['REEDS_DIR'] = reeds_dir
    ctx.obj['CF_YEARS'] = cf_years
    ctx.obj['BUILD_YEARS'] = build_years
    ctx.obj['SCENARIO'] = scenario
    ctx.obj['VERBOSE'] = verbose

    if ctx.invoked_subcommand is None:
        init_mult(name, out_dir, modules=['reVX', 'reV', 'rex'],
                  verbose=verbose)
        logger.info('Running reV to PLEXOS pipeline using job input: {}'
                    .format(job_input))
        logger.info('Outputs to be stored in: {}'.format(out_dir))
        logger.info('Aggregating plexos scenario "{}".'.format(scenario))
        for cf_year in cf_years:
            Manager.run(job_input, out_dir, reeds_dir, scenario=scenario,
                        cf_year=cf_year, build_years=build_years)


def get_node_cmd(name, job_input, out_dir, reeds_dir, cf_year, build_year,
                 scenario, verbose):
    """Get a CLI call command for the plexos CLI."""

    args = ['-n {}'.format(SLURM.s(name)),
            '-j {}'.format(SLURM.s(job_input)),
            '-o {}'.format(SLURM.s(out_dir)),
            '-rd {}'.format(SLURM.s(reeds_dir)),
            '-y [{}]'.format(SLURM.s(cf_year)),
            '-by [{}]'.format(SLURM.s(build_year)),
            '-s {}'.format(SLURM.s(scenario)),
            ]

    if verbose:
        args.append('-v')

    cmd = 'python -m reVX.plexos.plexos_cli {}'.format(' '.join(args))
    logger.debug('Creating the following command line call:\n\t{}'.format(cmd))

    return cmd


@main.command()
@click.option('--alloc', '-a', required=True, type=STR,
              help='Eagle allocation account name.')
@click.option('--memory', '-mem', default=90, type=INT, show_default=True,
              help='Eagle node memory request in GB. Default is 90')
@click.option('--walltime', '-wt', default=1.0, type=float, show_default=True,
              help='Eagle walltime request in hours. Default is 1.0')
@click.option('--feature', '-l', default=None, type=STR, show_default=True,
              help=('Additional flags for SLURM job. Format is "--qos=high" '
                    'or "--depend=[state:job_id]". Default is None.'))
@click.option('--stdout_path', '-sout', default=None, type=STR,
              show_default=True,
              help='Subprocess standard output path. Default is in out_dir.')
@click.pass_context
def eagle(ctx, alloc, memory, walltime, feature, stdout_path):
    """Eagle submission tool for PLEXOS aggregation."""

    name = ctx.obj['NAME']
    job_input = ctx.obj['JOB_INPUT']
    out_dir = ctx.obj['OUT_DIR']
    reeds_dir = ctx.obj['REEDS_DIR']
    cf_years = ctx.obj['CF_YEARS']
    build_years = ctx.obj['BUILD_YEARS']
    scenario = ctx.obj['SCENARIO']
    verbose = ctx.obj['VERBOSE']

    if stdout_path is None:
        stdout_path = os.path.join(out_dir, 'stdout/')

    if scenario is None:
        job = pd.read_csv(job_input)
        scenarios = job.scenario.unique()
    else:
        scenarios = [scenario]

    slurm_manager = ctx.obj.get('SLURM_MANAGER', None)
    if slurm_manager is None:
        slurm_manager = SLURM()
        ctx.obj['SLURM_MANAGER'] = slurm_manager

    for scenario in scenarios:
        for cf_year in cf_years:
            for build_year in build_years:
                node_name = ('{}_{}_{}_{}'
                             .format(name, scenario.replace(' ', '_'),
                                     build_year, cf_year))
                cmd = get_node_cmd(node_name, job_input, out_dir, reeds_dir,
                                   cf_year, build_year, scenario, verbose)

                logger.info('Running reVX plexos aggregation on Eagle with '
                            'node name "{}"'.format(node_name))

                out = slurm_manager.sbatch(cmd,
                                           alloc=alloc,
                                           memory=memory,
                                           walltime=walltime,
                                           feature=feature,
                                           name=node_name,
                                           stdout_path=stdout_path)[0]
                if out:
                    msg = ('Kicked off reVX plexos aggregation job "{}" '
                           '(SLURM jobid #{}) on Eagle.'
                           .format(node_name, out))
                else:
                    msg = ('Was unable to kick off reVX plexos aggregation job'
                           ' "{}". Please see the stdout error messages'
                           .format(node_name))

                click.echo(msg)
                logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running PLEXOS CLI')
        raise



# -*- coding: utf-8 -*-
"""
RPM Clustering Module
"""
from copy import deepcopy
import geopandas as gpd
import logging
import numpy as np
import pywt
from scipy.spatial import cKDTree
from shapely.geometry import Point

from reVX.handlers.outputs import Outputs
from reVX.utilities.cluster_methods import ClusteringMethods

logger = logging.getLogger(__name__)


class RPMClusters:
    """
    Base class for RPM clusters

    Examples
    --------
    >>> from reV import Resource
    >>>
    >>> fname = '$TESTDATADIR/reV_gen/gen_pv_2012.h5'
    >>> with Resource(fname) as res:
    >>>     gen_gids = f.meta.index.values
    >>>
    >>> clusters = RPMClusters(fname, gen_gids, n_clusters=6)
    >>> clusters._cluster(**kwargs)
    >>> clusters.meta
            gen_gid   latitude  longitude  cluster_id   geometry
    0         0  41.290001 -71.860001           0  POINT (-71.86000 41.29000)
    1         1  41.290001 -71.820000           0  POINT (-71.82000 41.29000)
    2         2  41.250000 -71.820000           4  POINT (-71.82000 41.25000)
    3         3  41.330002 -71.820000           0  POINT (-71.82000 41.33000)
    4         4  41.369999 -71.820000           0  POINT (-71.82000 41.37000)
    ..      ...        ...        ...         ...                         ...
    95       95  41.250000 -71.660004           4  POINT (-71.66000 41.25000)
    96       96  41.889999 -71.660004           5  POINT (-71.66000 41.89000)
    97       97  41.450001 -71.660004           3  POINT (-71.66000 41.45000)
    98       98  41.610001 -71.660004           1  POINT (-71.66000 41.61000)
    99       99  41.410000 -71.660004           3  POINT (-71.66000 41.41000)

    Generate Shape File of Cluster

    >>> RPMClusters._generate_shapefile(clusters.meta, fpath='./test.shp')
    """
    def __init__(self, cf_fpath, gen_gids, n_clusters):
        """
        Parameters
        ----------
        cf_fpath : str
            Path to reV .h5 files containing desired capacity factor profiles
        gen_gids : list | ndarray
            List or vector of gen_gids to cluster on
        n_clusters : int
            Number of clusters to identify
        """
        self._meta, self._coefficients = self._parse_data(cf_fpath,
                                                          gen_gids)
        self._n_clusters = n_clusters

    @property
    def coefficients(self):
        """
        Returns
        -------
        _coefficients : ndarray
            Array of wavelet coefficients for each gen_gid
        """
        return self._coefficients

    @property
    def meta(self):
        """
        Returns
        -------
        _meta : pandas.DataFrame
            DataFrame of meta data:
            - gen_gid
            - latitude
            - longitude
            - cluster_id
            - rank
        """
        return self._meta

    @property
    def n_clusters(self):
        """
        Returns
        -------
        _n_clusters : int
            Number of clusters
        """
        return self._n_clusters

    @property
    def cluster_coefficients(self):
        """
        Returns
        -------
        cluster_coeffs : ndarray
            Representative coefficients for each cluster
        """
        cluster_coeffs = None
        if 'cluster_id' in self._meta:
            cluster_coeffs = []
            for _, cdf in self._meta.groupby('cluster_id'):
                idx = cdf.index.values
                cluster_coeffs.append(self.coefficients[idx].mean(axis=0))

            cluster_coeffs = np.array(cluster_coeffs)

        return cluster_coeffs

    @property
    def cluster_ids(self):
        """
        Returns
        -------
        cluster_ids : ndarray
            Cluster cluster_id for each gen_gid
        """
        cluster_ids = None
        if 'cluster_id' in self._meta:
            cluster_ids = self._meta['cluster_id'].values
        return cluster_ids

    @property
    def cluster_coordinates(self):
        """
        Returns
        -------
        cluster_coords : ndarray
            lon, lat coordinates of the centroid of each cluster
        """
        cluster_coords = None
        if 'cluster_id' in self._meta:
            cluster_coords = self._meta.groupby('cluster_id')
            cluster_coords = cluster_coords[['longitude', 'latitude']].mean()
            cluster_coords = cluster_coords.values

        return cluster_coords

    @property
    def coordinates(self):
        """
        Returns
        -------
        coords : ndarray
            lon, lat coordinates for each gen_gid
        """
        coords = self._meta[['longitude', 'latitude']].values
        return coords

    @classmethod
    def _parse_data(cls, cf_fpath, gen_gids):
        """
        Extract lat, lon coordinates for given gen_gids
        Extract and convert cf_profiles into wavelet coefficients

        Parameters
        ----------
        cf_fpath : str
            Path to reV .h5 files containing desired capacity factor profiles
        gen_gids : list | ndarray
            List or vector of gen_gids to cluster on
        """

        with Outputs(cf_fpath, mode='r', unscale=False) as cfs:
            meta = cfs.meta.loc[gen_gids, ['latitude', 'longitude']]
            gid_slice, gid_idx = cls._gid_pos(gen_gids)
            coeff = cfs['cf_profile', :, gid_slice][:, gid_idx]

        meta['gen_gid'] = gen_gids
        cols = ['gen_gid', 'latitude', 'longitude']
        meta = meta[cols].reset_index(drop=True)
        coeff = cls._calculate_wavelets(coeff.T)
        return meta, coeff

    @staticmethod
    def _gid_pos(gen_gids):
        """
        Parameters
        ----------
        gen_gids : list | ndarray
            List or vector of gen_gids to cluster on

        Returns
        -------
        gid_slice : slice
            Slice that encompasses the entire gen_gid range
        gid_idx : ndarray
            Adjusted list to extract gen_gids of interest from slice
        """
        if isinstance(gen_gids, list):
            gen_gids = np.array(gen_gids)

        s = gen_gids.min()
        e = gen_gids.max() + 1
        gid_slice = slice(s, e, None)
        gid_idx = gen_gids - s

        return gid_slice, gid_idx

    @staticmethod
    def _calculate_wavelets(ts_arrays):
        """ Calculates the wavelet coefficients of each
            timeseries within ndarray """
        coefficients = RPMWavelets.get_dwt_coefficients(ts_arrays)
        return coefficients

    def _cluster_coefficients(self, method="kmeans", **kwargs):
        """ Apply a clustering method to <self.ts_arrays> """
        logger.debug('Applying {} clustering '.format(method))

        c_func = getattr(ClusteringMethods, method)
        labels = c_func(self.coefficients, n_clusters=self.n_clusters,
                        **kwargs)
        return labels

    def _dist_rank_optimization(self, norm=None):
        """
        Re-cluster data by minimizing the sum of the:
        - distance between each point and each cluster centroid
        - distance between each point and each

        Parameters
        ----------
        norm : str
            Normalization method to use (see sklearn.preprocessing.normalize)
            if None range normalize

        Returns
        -------
        new_labels : ndarray
            New cluster labels
        """
        cluster_coeffs = self.cluster_coefficients
        cluster_centroids = self.cluster_coordinates
        rmse = []
        dist = []
        for centroid, rep_coeffs in zip(cluster_centroids, cluster_coeffs):
            c_rmse = np.mean((self.coefficients - rep_coeffs) ** 2,
                             axis=1) ** 0.5
            rmse.append(c_rmse)
            c_dist = np.linalg.norm(self.coordinates - centroid, axis=1)
            dist.append(c_dist)

        rmse = ClusteringMethods._normalize_values(np.array(rmse), norm=norm)
        dist = ClusteringMethods._normalize_values(np.array(dist), norm=norm)
        err = (dist**2 + rmse**2)
        new_labels = np.argmin(err, axis=0)
        return new_labels

    def _dist_rank_filter(self, iterate=True, norm=None):
        """
        Re-cluster data by minimizing the sum of the:
        - distance between each point and each cluster centroid

        Parameters
        ----------
        iterate : bool
            Iterate on _dist_rank_optimization until cluster centroids and
            profiles start to converge
        norm : str
            Normalization method to use (see sklearn.preprocessing.normalize)
            if None range normalize

        Returns
        -------
        new_labels : ndarray
            New cluster labels
        """
        clusters = deepcopy(self)
        coeffs = clusters.cluster_coefficients
        centroids = clusters.cluster_coordinates
        dist, rmse = 0, 0
        while True:
            new_labels = clusters._dist_rank_optimization(norm=norm)
            clusters._meta['cluster_id'] = new_labels
            if iterate:
                c_coeffs = clusters.cluster_coefficients
                c_centroids = clusters.cluster_coordinates
                dist_i = np.linalg.norm(c_centroids - centroids)
                rmse_i = np.mean((c_coeffs - coeffs) ** 2) ** 0.5
                if (dist_i <= dist and rmse_i <= rmse):
                    break
                else:
                    dist, rmse = dist_i, rmse_i
                    coeffs, centroids = c_coeffs, c_centroids
            else:
                break

        return new_labels

    @staticmethod
    def _get_cluster_geom(gdf_points):
        """
        Generate cluster polygons as a geopandas dataframe
        """
        lookup = gdf_points[['latitude', 'longitude']]
        tree = cKDTree(lookup)  # pylint: disable=not-callable
        dists, _ = tree.query(lookup, k=2)
        mean_dist = dists.T[1].mean()
        gdf_poly = gdf_points.copy()
        gdf_poly.geometry = gdf_poly.geometry.buffer(mean_dist)
        clusters = gdf_poly.dissolve(by='cluster_id').reset_index()
        clusters.geometry = clusters.geometry.buffer(-mean_dist / 2)

        return clusters, mean_dist

    @classmethod
    def _generate_shapefile(cls, meta, fpath, beautify=True):
        """
        Generate cluster polygons and save to shapefile
        """
        geometry = [Point(xy) for xy in zip(meta.longitude, meta.latitude)]
        gdf_points = gpd.GeoDataFrame(meta, geometry=geometry,
                                      crs="EPSG:5070")

        clusters, mean_dist = cls._get_cluster_geom(gdf_points)

        if beautify:
            clusters.geometry = clusters.geometry.buffer(-mean_dist)
            clusters.geometry = clusters.geometry.buffer(mean_dist)
            for index, _ in clusters.iterrows():
                geom = clusters.loc[index, 'geometry']
                if geom.geom_type == 'MultiPolygon':
                    clusters.loc[index, 'geometry'] = max(geom,
                                                          key=lambda a: a.area)

        clusters[['cluster_id', 'geometry']].to_file(fpath)
        return fpath

    def _contiguous_filter(self, drop_islands=True, buffer_weight=2):
        """
        Re-classify clusters by making contigous cluster polygons
        """
        meta = self._meta

        geometry = [Point(xy) for xy in zip(meta.longitude, meta.latitude)]
        gdf_points = gpd.GeoDataFrame(meta, geometry=geometry,
                                      crs="EPSG:5070")

        clusters, mean_dist = self._get_cluster_geom(gdf_points)

        # Drop Islands
        if drop_islands:
            buffer = buffer_weight * mean_dist
            clusters.geometry = clusters.geometry.buffer(-buffer)
            clusters.geometry = clusters.geometry.buffer(buffer)
            for index, _ in clusters.iterrows():
                geom = clusters.loc[index, 'geometry']
                if geom.geom_type == 'MultiPolygon':
                    clusters.loc[index, 'geometry'] = max(geom,
                                                          key=lambda a: a.area)

        intersected = gpd.sjoin(gdf_points, clusters,
                                how="left", op='intersects')

        # drop duplicate rows
        gid_counts = intersected.groupby('gen_gid_left').size()
        duplicate_gids = gid_counts[gid_counts > 1].index
        mask = intersected['gen_gid_left'].isin(duplicate_gids)
        intersected.loc[mask, 'cluster_id_right'] = None
        intersected = intersected.drop_duplicates(subset=['gen_gid_left'])

        mask = np.isnan(intersected.cluster_id_right)
        assigned = intersected[~mask].reset_index()
        unassigned = intersected[mask]

        lookup = assigned[['latitude_left', 'longitude_left']]
        target = unassigned[['latitude_left', 'longitude_left']]
        tree = cKDTree(lookup)  # pylint: disable=not-callable
        _, inds = tree.query(target, k=1)

        for i, ind in enumerate(list(unassigned.index)):
            nearest_cluster_id = assigned.loc[inds[i], 'cluster_id_left']
            intersected.loc[ind, 'cluster_id_left'] = nearest_cluster_id

        new_labels = intersected['cluster_id_left']

        return new_labels

    def _calculate_ranks(self):
        """ Determine the rank of each location within all clusters
        based on the mean square errors """
        cluster_coeffs = self.cluster_coefficients
        for i, cdf in self.meta.groupby('cluster_id'):
            pos = cdf.index
            rep_coeffs = cluster_coeffs[i]
            coeffs = self.coefficients[pos]
            err = np.mean((coeffs - rep_coeffs) ** 2, axis=1) ** 0.5
            rank = np.argsort(err)
            self._meta.loc[pos, 'rank'] = rank

    def _cluster(self, method='kmeans', method_kwargs=None,
                 dist_rank_filter=True, dist_rmse_kwargs=None,
                 contiguous_filter=True, contiguous_kwargs=None):
        """
        Run three step RPM clustering procedure:
        1) Cluster on wavelet coefficients
        2) Clean up clusters by optimizing rmse and distance
        3) Remove islands using polygon intersection

        Parameters
        ----------
        method : str
            Method to use to cluster coefficients
        method_kwargs : dict
            Kwargs for running _cluster_coefficients
        dist_rank_filter : bool
            Run _optimize_dist_rank
        dist_rmse_kwargs : dict
            Kwargs for running _dist_rank_optimization
        contiguous_filter : bool
            Run _contiguous_filter
        contiguous_kwargs : dict
            Kwargs for _contiguous_filter
        """

        if self.n_clusters <= 1:
            dist_rank_filter = False
            contiguous_filter = False

        if method_kwargs is None:
            method_kwargs = {}

        labels = self._cluster_coefficients(method=method, **method_kwargs)
        self._meta['cluster_id'] = labels

        # Optimize Distance & Rank
        if dist_rank_filter is True:
            if dist_rmse_kwargs is None:
                dist_rmse_kwargs = {}

            new_labels = self._dist_rank_filter(**dist_rmse_kwargs)
            self._meta['cluster_id'] = new_labels

        # Apply contiguous filter
        if contiguous_filter is True:
            if contiguous_kwargs is None:
                contiguous_kwargs = {}

            new_labels = self._contiguous_filter(**contiguous_kwargs)
            self._meta['cluster_id'] = new_labels

        self._calculate_ranks()

    @classmethod
    def cluster(cls, cf_h5_path, region_gen_gids, n_clusters, method='kmeans',
                method_kwargs=None, dist_rank_filter=True,
                dist_rmse_kwargs=None, contiguous_filter=True,
                contiguous_kwargs=None):
        """
        Entry point for RPMCluster to get clusters for a given region
        defined as a list | array of gen_gids

        Parameters
        ----------
        cf_h5_path : str
            Path to reV .h5 files containing desired capacity factor profiles
        region_gen_gids : list | ndarray
            List or vector of gen_gids to cluster on
        n_clusters : int
            Number of clusters to identify
        method : str
            Method to use to cluster coefficients
        method_kwargs : dict
            Kwargs for running _cluster_coefficients
        dist_rank_filter : bool
            Re-cluster data by minimizing the sum of the:
            - distance between each point and each cluster centroid
        dist_rmse_kwargs : dict
            Kwargs for running _dist_rank_optimization
        contiguous_filter : bool
            Re-classify clusters by making contigous cluster polygons
        contiguous_kwargs : dict
            Kwargs for _contiguous_filter

        Returns
        -------
        out : pandas.DataFrame
            Cluster results: (gen_gid, lon, lat, cluster_id, rank)

        Examples
        --------
        >>> from reV import Resource
        >>>
        >>> fname = '$TESTDATADIR/reV_ge/gen_pv_2012.h5'
        >>> with Resource(fname) as res:
        >>>     gen_gids = f.meta.index.values
        >>>
        >>> RPMClusters.cluster(fname, gen_gids, n_clusters=6)
                gen_gid   latitude  longitude  cluster_id   geometry
        0         0  41.290001 -71.860001       0  POINT (-71.86000 41.29000)
        1         1  41.290001 -71.820000       0  POINT (-71.82000 41.29000)
        2         2  41.250000 -71.820000       4  POINT (-71.82000 41.25000)
        3         3  41.330002 -71.820000       0  POINT (-71.82000 41.33000)
        4         4  41.369999 -71.820000       0  POINT (-71.82000 41.37000)
        ..      ...        ...        ...     ...                         ...
        95       95  41.250000 -71.660004       4  POINT (-71.66000 41.25000)
        96       96  41.889999 -71.660004       5  POINT (-71.66000 41.89000)
        97       97  41.450001 -71.660004       3  POINT (-71.66000 41.45000)
        98       98  41.610001 -71.660004       1  POINT (-71.66000 41.61000)
        99       99  41.410000 -71.660004       3  POINT (-71.66000 41.41000)
        """
        clusters = cls(cf_h5_path, region_gen_gids, n_clusters)
        try:
            clusters._cluster(method=method, method_kwargs=method_kwargs,
                              dist_rank_filter=dist_rank_filter,
                              dist_rmse_kwargs=dist_rmse_kwargs,
                              contiguous_filter=contiguous_filter,
                              contiguous_kwargs=contiguous_kwargs)
        except Exception as e:
            logger.exception('Clustering failed on gen_gids {} through {}: {}'
                             .format(np.min(region_gen_gids),
                                     np.max(region_gen_gids), e))
        return clusters.meta


class RPMWavelets:
    """Base class for RPM wavelets"""

    @classmethod
    def get_dwt_coefficients(cls, x, wavelet='Haar', level=None, indices=None):
        """
        Collect wavelet coefficients for time series <x> using
        mother wavelet <wavelet> at levels <level>.

        Parameters
        ----------
        x : ndarray
            time series values
        wavelet : string
            mother wavelet type
        level : int
            optional wavelet computation level
        indices : ndarray
            coefficient array levels to keep

        Returns
        -------
        list
            stacked coefficients at <indices>
        """

        # set mother
        _wavelet = pywt.Wavelet(wavelet)

        # multi-level with default depth
        logger.debug('Calculating wavelet coefficients'
                     ' with {w} wavelet'.format(w=_wavelet.family_name))

        _wavedec = pywt.wavedec(data=x, wavelet=_wavelet, axis=1, level=level)

        return cls._subset_coefficients(x=_wavedec,
                                        gid_count=x.shape[0],
                                        indices=indices)

    @staticmethod
    def _subset_coefficients(x, gid_count, indices=None):
        """
        Subset and stack wavelet coefficients

        Parameters
        ----------
        x : ndarray
            coefficients arrays
        gid_count : int
            number of area ID values
        indices : ndarray
            coefficient array levels to keep

        Returns
        -------
        ndarray
            stacked coefficients: converted to integers
        """

        indices = indices or range(0, len(x))

        _coefficient_count = 0
        for _index in indices:
            _shape = x[_index].shape
            _coefficient_count += _shape[1]

        _combined_wc = np.empty(shape=(gid_count, _coefficient_count),
                                dtype=np.int32)

        logger.debug('{c:d} coefficients'.format(c=_coefficient_count))

        _i_start = 0
        for _index in indices:
            _i_end = _i_start + x[_index].shape[1]
            _combined_wc[:, _i_start:_i_end] = np.round(x[_index])
            _i_start = _i_end

        return _combined_wc



# -*- coding: utf-8 -*-
"""
RPM Command Line Interface
"""
import click
import logging
import os
from rex.utilities.loggers import init_mult
from rex.utilities.cli_dtypes import STR, INT
from rex.utilities.hpc import SLURM
from rex.utilities.utilities import dict_str_load, get_class_properties

from reVX.config.rpm import (RPMConfig, ClusterConfigGroup,
                             RepProfilesConfigGroup)
from reVX.rpm.rpm_manager import RPMClusterManager as rpm_cm
from reVX.rpm.rpm_output import RPMOutput as rpm_o
from reVX.utilities.exceptions import RPMRuntimeError
from reVX import __version__

logger = logging.getLogger(__name__)


@click.group()
@click.option('--name', '-n', default='RPM', type=STR,
              show_default=True,
              help='Job name. Default is "RPM".')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def main(ctx, name, verbose):
    """
    RPM Command Line Interface
    """
    ctx.ensure_object(dict)
    ctx.obj['VERBOSE'] = verbose
    ctx.obj['NAME'] = name


@main.command()
def version():
    """
    print version
    """
    click.echo(__version__)


@main.command()
def valid_config_keys():
    """
    Echo the valid RPM config keys
    """
    config_classes = [RPMConfig, ClusterConfigGroup, RepProfilesConfigGroup]
    for cls in config_classes:
        cls_name = str(cls).split('.')[-1].strip("'>")
        click.echo("Valid keys for {}: {}"
                   .format(cls_name, ', '.join(get_class_properties(cls))))


def run_local(ctx, config):
    """
    Run reV to ReEDs locally from config

    Parameters
    ----------
    ctx : click.ctx
        click ctx object
    config : RPMConfig
        RPM Config object
    """
    ctx.obj['NAME'] = config.name
    ctx.invoke(local,
               out_dir=config.dirout,
               cf_profiles=config.cf_profiles,
               log_dir=config.logdir,
               max_workers=config.max_workers)

    if config.cluster is not None:
        ctx.invoke(cluster,
                   rpm_meta=config.cluster.rpm_meta,
                   region_col=config.cluster.region_col,
                   dist_rank_filter=config.cluster.dist_rank_filter,
                   contiguous_filter=config.cluster.contiguous_filter)

    if config.rep_profiles is not None:
        ctx.invoke(rep_profiles,
                   rpm_clusters=config.rep_profiles.rpm_clusters,
                   exclusions=config.rep_profiles.exclusions,
                   excl_dict=config.rep_profiles.excl_dict,
                   techmap_dset=config.rep_profiles.techmap_dset,
                   trg=config.rep_profiles.trg,
                   n_profiles=config.rep_profiles.n_profiles,
                   forecast_fpath=config.rep_profiles.forecast_fpath)


@main.command()
@click.option('--config', '-c', required=True,
              type=click.Path(exists=True),
              help='Filepath to reVX-rpm config json file.')
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def from_config(ctx, config, verbose):
    """
    Run reVX-rpm from a config.
    """

    config = RPMConfig(config)

    if 'VERBOSE' in ctx.obj:
        if any((ctx.obj['VERBOSE'], verbose)):
            config._log_level = logging.DEBUG
    elif verbose:
        config._log_level = logging.DEBUG

    if config.execution_control.option == 'local':
        run_local(ctx, config)
    elif config.execution_control.option == 'eagle':
        eagle(config)


@main.group(chain=True)
@click.option('--out_dir', '-o', required=True, type=click.Path(),
              help='Directory to dump output files')
@click.option('--cf_profiles', '-cf', required=True,
              type=click.Path(exists=True),
              help=('Path to reV .h5 file containing desired capacity factor '
                    'profiles'))
@click.option('--log_dir', '-log', default=None, type=STR, show_default=True,
              help='Directory to dump log files. Default is out_dir.')
@click.option('--max_workers', '-mw', type=INT, default=None,
              show_default=True,
              help=('Number of parallel workers. 1 will run serial, '
                    'None will use all available.'))
@click.option('--verbose', '-v', is_flag=True,
              help='Flag to turn on debug logging. Default is not verbose.')
@click.pass_context
def local(ctx, out_dir, cf_profiles, log_dir, max_workers, verbose):
    """
    Run reVX-REEDS on local hardware.
    """
    ctx.obj['OUT_DIR'] = out_dir
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)

    if log_dir is None:
        log_dir = out_dir

    name = ctx.obj['NAME']
    if 'VERBOSE' in ctx.obj:
        verbose = any((ctx.obj['VERBOSE'], verbose))

    log_modules = ['reVX', 'reV', 'rex']
    init_mult(name, log_dir, modules=log_modules, verbose=verbose)

    logger.info('Running reV to RPM pipeline\n'
                'Outputs to be stored in: {}'.format(out_dir))

    ctx.obj['CF_PROFILES'] = cf_profiles
    ctx.obj['MAX_WORKERS'] = max_workers


@local.command()
@click.option('--rpm_meta', '-m', required=True, type=click.Path(exists=True),
              help='.csv or .json containing the RPM meta data')
@click.option('--region_col', '-reg', type=str, default=None,
              show_default=True,
              help='The meta-data field to map RPM regions to')
@click.option('--dist_rank_filter', '-drf', is_flag=True,
              help=('Re-cluster data by minimizing the sum of the: '
                    'distance between each point and each cluster centroid'))
@click.option('--contiguous_filter', '-cf', is_flag=True,
              help=('Flag to re-classify clusters by making contigous cluster '
                    'polygons'))
@click.pass_context
def cluster(ctx, rpm_meta, region_col, dist_rank_filter, contiguous_filter):
    """
    Cluster RPM Regions
    """
    name = ctx.obj['NAME']
    cf_profiles = ctx.obj['CF_PROFILES']
    out_dir = ctx.obj['OUT_DIR']
    max_workers = ctx.obj['MAX_WORKERS']

    logger.info('Clustering regions based on:\n{}'.format(rpm_meta))
    rpm_clusters = rpm_cm.run_clusters(cf_profiles, rpm_meta, out_dir,
                                       job_tag=name, rpm_region_col=region_col,
                                       max_workers=max_workers,
                                       dist_rank_filter=dist_rank_filter,
                                       contiguous_filter=contiguous_filter)

    logger.info('reVX - RPM clustering methods complete.')
    ctx.obj['RPM_CLUSTERS'] = rpm_clusters


@local.command()
@click.option('--rpm_clusters', '-rc', type=STR, default=None,
              help=('Path to pre-existing RPM cluster results .csv with '
                    '(gid, gen_gid, cluster_id, rank)'))
@click.option('--exclusions', '-excl', default=None,
              type=click.Path(exists=True), show_default=True,
              help=('Filepath to exclusions data (must match the techmap grid)'
                    ' None will not apply exclusions.'))
@click.option('--excl_dict', '-exd', default=None, type=STR, show_default=True,
              help='String representation of a dictionary of exclusion '
              'LayerMask arguments {layer: {kwarg: value}} where layer is a '
              'dataset in excl_fpath and kwarg can be "inclusion_range", '
              '"exclude_values", "include_values", "use_as_weights", '
              'or "weight".')
@click.option('--techmap_dset', '-tmd', default=None, type=STR,
              show_default=True,
              help=('Dataset name in the techmap file containing the '
                    'exclusions-to-resource mapping data.'))
@click.option('--trg', '-trg', default=None, type=STR, show_default=True,
              help=('Filepath to TRG LCOE bins.'))
@click.option('--n_profiles', '-np', type=INT, default=1, show_default=True,
              help=('Number of profiles per cluster to export.'))
@click.option('--forecast_fpath', '-fcst', type=STR, default=None,
              show_default=True,
              help=('reV generation output file for forecast data. If this is '
                    'input, profiles will be taken from forecast file instead '
                    'of the cf file, based on a NN mapping.'))
@click.pass_context
def rep_profiles(ctx, rpm_clusters, exclusions, excl_dict, techmap_dset, trg,
                 n_profiles, forecast_fpath):
    """
    Extract representative profiles from RPM clusters
    """
    name = ctx.obj['NAME']
    cf_profiles = ctx.obj['CF_PROFILES']
    out_dir = ctx.obj['OUT_DIR']
    max_workers = ctx.obj['MAX_WORKERS']
    if rpm_clusters is None:
        if 'RPM_CLUSTERS' not in ctx.obj:
            msg = ('You must run "cluster" or provide path to existing '
                   'RPM clusters to extract representative profiles!')
            logger.error(msg)
            raise RPMRuntimeError(msg)

        rpm_clusters = ctx.obj['RPM_CLUSTERS']

    logger.info('Extracting representative profiles from RPM clusters: {}'
                .format(rpm_clusters))
    logger.info('Extracting representative profiles using exclusions: {}'
                .format(exclusions))

    if trg is not None:
        logger.info('Applying TRGs: {}'.format(trg))

    if isinstance(excl_dict, str):
        excl_dict = dict_str_load(excl_dict)

    rpm_o.process_outputs(rpm_clusters, cf_profiles, exclusions, excl_dict,
                          techmap_dset, out_dir, job_tag=name,
                          max_workers=max_workers, trg=trg)

    if forecast_fpath is not None or n_profiles > 1:
        logger.info('Extracting extra representative profiles from: {}'
                    .format(rpm_clusters))
        if forecast_fpath is not None:
            logger.info('Using forecast file: {}'.format(forecast_fpath))

        rpm_o.extract_profiles(rpm_clusters, cf_profiles, out_dir,
                               n_profiles=n_profiles, job_tag=name,
                               max_workers=max_workers,
                               forecast_fpath=forecast_fpath)


def get_node_cmd(config):
    """
    Get the node CLI call for the reVX-REEDS pipeline.

    Parameters
    ----------
    config : reVX.config.reeds.ReedsConfig
        reVX-REEDS config object.

    Returns
    -------
    cmd : str
        CLI call to submit to SLURM execution.
    """

    args = ['-n {}'.format(SLURM.s(config.name)),
            'local',
            '-o {}'.format(SLURM.s(config.dirout)),
            '-cf {}'.format(SLURM.s(config.cf_profiles)),
            '-mw {}'.format(SLURM.s(config.max_workers)),
            '-log {}'.format(SLURM.s(config.logdir)),
            ]

    if config.log_level == logging.DEBUG:
        args.append('-v')

    if config.cluster is not None:
        cluster = ['cluster',
                   '-m {}'.format(SLURM.s(config.cluster.rpm_meta)),
                   '-reg {}'.format(SLURM.s(config.cluster.region_col)),
                   ]

        if config.cluster.dist_rank_filter:
            cluster.append('-drf')

        if config.cluster.contiguous_filter:
            cluster.append('-cf')

        args.extend(cluster)

    if config.rep_profiles is not None:
        rep_profiles = config.rep_profiles.copy()
        profiles = ['rep_profiles',
                    '-rc {}'.format(SLURM.s(rep_profiles.rpm_clusters)),
                    '-excl {}'.format(SLURM.s(rep_profiles.exclusions)),
                    '-exd {}'.format(SLURM.s(rep_profiles.excl_dict)),
                    '-tmp {}'.format(SLURM.s(rep_profiles.techmap_dset)),
                    '-trg {}'.format(SLURM.s(rep_profiles.trg)),
                    '-np {}'.format(SLURM.s(rep_profiles.n_profiles)),
                    '-fcst {}'.format(SLURM.s(rep_profiles.forecast_fpath)),
                    ]

        args.extend(profiles)

    cmd = 'python -m reVX.rpm.rpm_cli {}'.format(' '.join(args))
    logger.debug('Submitting the following cli call:\n\t{}'.format(cmd))

    return cmd


def eagle(config):
    """
    Run reVX-RPM on Eagle HPC.

    Parameters
    ----------
    config : reVX.config.rpm.RPMConfig
        reVX-RPM config object.
    """

    cmd = get_node_cmd(config)
    name = config.name
    log_dir = config.logdir
    stdout_path = os.path.join(log_dir, 'stdout/')

    logger.info('Running reVX-RPM pipeline on Eagle with '
                'node name "{}"'.format(name))
    slurm_manager = SLURM()
    out = slurm_manager.sbatch(cmd,
                               name=name,
                               stdout_path=stdout_path,
                               alloc=config.execution_control.alloc,
                               memory=config.execution_control.node_mem,
                               walltime=config.execution_control.walltime,
                               feature=config.execution_control.feature,
                               module=config.execution_control.module,
                               conda_env=config.execution_control.conda_env,
                               )[0]
    if out:
        msg = ('Kicked off reVX-RPM pipeline job "{}" '
               '(SLURM jobid #{}) on Eagle.'
               .format(name, out))
    else:
        msg = ('Was unable to kick off reVX-RPM pipeline job "{}". '
               'Please see the stdout error messages'
               .format(name))

    click.echo(msg)
    logger.info(msg)


if __name__ == '__main__':
    try:
        main(obj={})
    except Exception:
        logger.exception('Error running RPM CLI')
        raise



# -*- coding: utf-8 -*-
"""
Aggregate powerrose and sort directions by dominance
"""
import logging

from reV.supply_curve.aggregation import Aggregation

logger = logging.getLogger(__name__)


class MeanWindDirections(Aggregation):
    """
    Average the wind direction via the wind vectors.
    Then convert to equivalent sc_point_gid
    """

    def __init__(self, res_h5_fpath, excl_fpath, wdir_dsets,
                 tm_dset='techmap_wtk', excl_dict=None,
                 area_filter_kernel='queen', min_area=None,
                 check_excl_layers=False, resolution=128, excl_area=None):
        """
        Parameters
        ----------
        res_h5_fpath : str
            Filepath to .h5 file containing wind direction data
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        wdir_dsets : str | list
            Wind direction dataset to average
        tm_dset : str, optional
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data,
            by default 'techmap_wtk'
        excl_dict : dict | None, optional
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default 'queen'
        min_area : float | None, optional
            Minimum required contiguous area filter in sq-km, by default None
        check_excl_layers : bool, optional
            Run a pre-flight check on each exclusion layer to ensure they
            contain un-excluded values, by default False
        resolution : int | None, optional
            SC resolution, must be input in combination with gid,
            by default 128
        excl_area : float | None, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        """
        if isinstance(wdir_dsets, str):
            wdir_dsets = [wdir_dsets]

        for dset in wdir_dsets:
            if not dset.startswith('winddirection'):
                msg = ('{} is not a valid wind direction dataset!'
                       .format(dset))
                logger.error(msg)
                raise ValueError(msg)

        super().__init__(excl_fpath, res_h5_fpath, tm_dset, *wdir_dsets,
                         excl_dict=excl_dict,
                         area_filter_kernel=area_filter_kernel,
                         min_area=min_area,
                         check_excl_layers=check_excl_layers,
                         resolution=resolution, excl_area=excl_area)

    def aggregate(self, max_workers=None, chunk_point_len=1000):
        """
        Average wind directions to sc_points

        Parameters
        ----------
        max_workers : int | None
            Number of cores to run summary on. None is all
            available cpus.
        chunk_point_len : int
            Number of SC points to process on a single parallel worker.

        Returns
        -------
        agg : dict
            Aggregated values for each aggregation dataset
        """
        agg = super().aggregate(agg_method='mean_wind_dir',
                                max_workers=max_workers,
                                chunk_point_len=chunk_point_len)

        return agg

    @classmethod
    def run(cls, res_h5_fpath, excl_fpath, wdir_dsets,
            tm_dset='techmap_wtk', excl_dict=None,
            area_filter_kernel='queen', min_area=None,
            check_excl_layers=False, resolution=128, excl_area=None,
            max_workers=None, chunk_point_len=1000, out_fpath=None):
        """
        Aggregate powerrose to supply curve points, find neighboring supply
        curve point gids and rank them based on prominent powerrose direction

        Parameters
        ----------
        res_h5_fpath : str
            Filepath to .h5 file containing wind direction data
        excl_fpath : str
            Filepath to exclusions h5 with techmap dataset.
        wdir_dsets : str | list
            Wind direction dataset to average
        tm_dset : str, optional
            Dataset name in the techmap file containing the
            exclusions-to-resource mapping data,
            by default 'techmap_wtk'
        excl_dict : dict | None, optional
            Dictionary of exclusion LayerMask arugments {layer: {kwarg: value}}
            by default None
        area_filter_kernel : str, optional
            Contiguous area filter method to use on final exclusions mask,
            by default 'queen'
        min_area : float | None, optional
            Minimum required contiguous area filter in sq-km, by default None
        check_excl_layers : bool, optional
            Run a pre-flight check on each exclusion layer to ensure they
            contain un-excluded values, by default False
        resolution : int | None, optional
            SC resolution, must be input in combination with gid,
            by default 128
        excl_area : float | None, optional
            Area of an exclusion pixel in km2. None will try to infer the area
            from the profile transform attribute in excl_fpath,
            by default None
        max_workers : int | None, optional
            Number of cores to run summary on. None is all
            available cpus, by default None
        chunk_point_len : int, optional
            Number of SC points to process on a single parallel worker,
            by default 100
        out_fpath : str
            Path to .h5 file to save aggregated data too

        Returns
        -------
        agg : dict
            Aggregated values for each aggregation dataset
        """
        wdir = cls(res_h5_fpath, excl_fpath, wdir_dsets, tm_dset=tm_dset,
                   excl_dict=excl_dict,
                   area_filter_kernel=area_filter_kernel,
                   min_area=min_area,
                   check_excl_layers=check_excl_layers,
                   resolution=resolution, excl_area=excl_area)

        agg = wdir.aggregate(max_workers=max_workers,
                             chunk_point_len=chunk_point_len)

        if out_fpath is not None:
            wdir.save_agg_to_h5(out_fpath, agg)

        return agg



# -*- coding: utf-8 -*-
"""
reVX ProminentWindDirections Configuration
"""

from reV.config.base_analysis_config import AnalysisConfig


class ProminentWindDirsConfig(AnalysisConfig):
    """Config framework for prominent wind direction calculation"""

    NAME = 'ProminentWindDirs'
    REQUIREMENTS = ('powerrose_h5_fpath', 'excl_fpath')

    def __init__(self, config):
        """
        Parameters
        ----------
        config : dict
            Dictionary with pre-extracted config input group.
        """
        super().__init__(config)
        self._default_agg_dset = 'powerrose_100m'
        self._default_tm_dset = 'techmap_wtk'
        self._default_resolution = 128
        self._default_chunk_point_len = 1000

    @property
    def powerrose_h5_fpath(self):
        """Get the powerrose .h5 file path (required)."""
        return self['powerrose_h5_fpath']

    @property
    def excl_fpath(self):
        """Get the exclusions .h5 file path (required)."""
        return self['excl_fpath']

    @property
    def agg_dset(self):
        """Get the aggregation dataset name."""
        return self.get('agg_dset', self._default_agg_dset)

    @property
    def tm_dset(self):
        """Get the techmap dataset name."""
        return self.get('tm_dset', self._default_tm_dset)

    @property
    def resolution(self):
        """Get the supply curve resolution."""
        return self.get('resolution', self._default_resolution)

    @property
    def excl_area(self):
        """Get the exclusion pixel area in km2"""
        return self.get('excl_area', None)

    @property
    def max_workers(self):
        """Get the maximum number of workers."""
        return self.get('max_workers', None)

    @property
    def chunk_point_len(self):
        """Get the chunk length for parallel computation."""
        return self.get('chunk_point_len', self._default_chunk_point_len)



"""
Forecast Utilities
"""
import h5py
import logging
import numpy as np
import os
import shutil
from warnings import warn

from rex import Resource

logger = logging.getLogger(__name__)


class Forecasts:
    """
    Forecast Utility
    """
    def __init__(self, fcst_h5, fcst_dset, actuals_h5=None, actuals_dset=None):
        """
        Parameters
        ----------
        fcst_h5 : str
            Path to forecast .h5 file
        fcst_dset : str
            Dataset to correct
        actuals_h5 : str, optional
            Path to forecast to .h5 file, by default None
        actuals_dset : str, optional
            Actuals dataset, by default None
        """
        self._fcst_h5 = fcst_h5
        if actuals_h5 is None:
            actuals_h5 = fcst_h5

        self._actuals_h5 = actuals_h5

        self._fcst_dset = fcst_dset

        if actuals_dset is None:
            actuals_dset = fcst_dset

        self._actuals_dset = actuals_dset

        self._a_slice = self._preflight_check()

    @property
    def fcst_h5(self):
        """
        Path to forecast .h5 file

        Returns
        -------
        str
        """
        return self._fcst_h5

    @property
    def actuals_h5(self):
        """
        Path to actuals .h5 file

        Returns
        -------
        str
        """
        return self._actuals_h5

    @property
    def fcst_dset(self):
        """
        Forecast dset to correct

        Returns
        -------
        str
        """
        return self._fcst_dset

    @property
    def actuals_dset(self):
        """
        Actuals dset to use to correct forecasts

        Returns
        -------
        str
        """
        return self._actuals_dset

    @staticmethod
    def bias_correct_fcst(actuals, fcsts):
        """
        Bias correct forecast data using bias correction factor:
        total actual generation / total forecasted generation

        Parameters
        ----------
        actuals : ndarray
            Timeseries actuals (time x sites)
        fcsts : ndarray
            Timeseries forecats (time x sites)

        Returns
        -------
        fcsts : ndarray
            Bias corrected forecasts
        """
        bc_factors = actuals.sum(axis=0) / fcsts.sum(axis=0)
        fcsts = fcsts * bc_factors
        actuals_max = actuals.max(axis=0)
        mask = fcsts >= actuals_max
        fcsts = np.where(mask, actuals_max, fcsts)

        return fcsts

    @classmethod
    def blend_fcsts(cls, actuals, fcsts, fcst_perc):
        """
        Bias correct and blend forecasts with actuals

        Parameters
        ----------
        actuals : ndarray
            Timeseries actuals (time x sites)
        fcsts : ndarray
            Timeseries forecats (time x sites)
        fcst_perc : float
            Percentage of forecast to use for blending

        Returns
        -------
        fcsts : ndarray
            Bias corrected and blended forecasts
        """
        fcsts = cls.bias_correct_fcst(actuals, fcsts)
        fcsts = ((actuals * (1 - fcst_perc)) + (fcsts * fcst_perc))

        return fcsts

    @classmethod
    def _correct(cls, fcst, actuals, fcst_perc=None):
        """
        Correct given data using following steps:
        - Bias correct forecast data using bias correction factor:
        total actual generation / total forecasted generation
        - Blend fcst_perc of forecast generation with (1 - fcst_perc) of
        actuals generation

        Parameters
        ----------
        actuals : ndarray
            Timeseries actuals (time x sites)
        fcsts : ndarray
            Timeseries forecats (time x sites)
        fcst_perc : float, optional
            Percentage of forecast to use for blending, by default None

        Returns
        -------
        fcsts : ndarray
            Corrected forecasts
        """
        mae = cls.compute_mae(actuals, fcst)
        logger.debug('Forecast agg MAE: {:.4f}, ave MAE {:.4f}'
                     .format(*mae))

        if fcst_perc is not None:
            logger.info('Bias correcting and blending forecasts with '
                        '{:}% actuals'.format((1 - fcst_perc) * 100))
            fcst = cls.blend_fcsts(actuals, fcst, fcst_perc)
        else:
            logger.info('Bias correcting forecasts')
            fcst = cls.bias_correct_fcst(actuals, fcst)

        mae = cls.compute_mae(actuals, fcst)
        logger.debug('Corrected forecast agg MAE: {:.4f}, '
                     'ave MAE {:.4f}'.format(*mae))

        return fcst

    @staticmethod
    def compute_mae(actuals, fcsts):
        """
        Compute aggregate and normalized average MAE between actuals and
        forecasts. The MAE is normalized by the total actuals generation

        Parameters
        ----------
        actuals : ndarray
            Timeseries actuals (time x sites)
        fcsts : ndarray
            Timeseries forecats (time x sites)

        Returns
        -------
        agg_mae : float
            Aggregate MAE for all sites
        ave_mae : float
            Average MEA for all sites
        """
        site_mae = np.mean(np.abs(fcsts - actuals), axis=0)
        site_rel_mae = site_mae / np.max(actuals, axis=0)
        ave_mae = np.nanmean(site_rel_mae)
        agg_mae = np.nansum(site_rel_mae)

        return agg_mae, ave_mae

    @staticmethod
    def _make_time_index_name(dset_name):
        """
        Make time_index name from dataset name

        Parameters
        ----------
        dset_name : str
            profiles dataset

        Returns
        -------
        str
            time_index name associated with profiles dataset
        """
        time_index_name = 'time_index'
        if '/' in dset_name:
            group = dset_name.split('/')[0]
            time_index_name = "{}/{}".format(group, time_index_name)

        return time_index_name

    def _get_actuals_slice(self):
        """
        Create actuals slice needed to match fcst time-steps

        Returns
        -------
        a_slice : slice
            Slice of actuals timeseries needed to match forecast timeseries
        """
        with Resource(self.fcst_h5) as f:
            f_shape = f.get_dset_properties(self.fcst_dset)[0]
            f_time_index = f[self._make_time_index_name(self.fcst_dset)]

        with Resource(self.actuals_h5) as f:
            a_shape = f.get_dset_properties(self.actuals_dset)[0]
            a_time_index = f[self._make_time_index_name(self.actuals_dset)]

        a_slice = slice(None, None, a_shape[0] // f_shape[0])
        if len(f_time_index) != f_shape[0]:
            msg = ('Forecast time_index does not match forecast shape, '
                   'actuals slice will be estimated!')
            logger.warning(msg)
            warn(msg)
        elif not f_time_index.equals(a_time_index[a_slice]):
            a_slice = np.where(a_time_index.isin(f_time_index))[0]

        return a_slice

    def _preflight_check(self):
        """
        Check to ensure dset is available in forecast and actuals .h5 files

        Returns
        -------
        a_slice : slice
            Slice of actuals timeseries needed to match forecast timeseries
        """
        with Resource(self.fcst_h5) as f:
            if self.fcst_dset not in f:
                msg = ('{} is not a valid dataset in forecast file: {}'
                       .format(self.fcst_dset, self.fcst_h5))
                logger.error(msg)
                raise RuntimeError(msg)
            else:
                f_shape = f.get_dset_properties(self.fcst_dset)[0]

        with Resource(self.actuals_h5) as f:
            if self.actuals_dset not in f:
                msg = ('{} is not a valid dataset in actuals file: {}'
                       .format(self.actuals_dset, self.actuals_h5))
                logger.error(msg)
                raise RuntimeError(msg)
            else:
                a_shape = f.get_dset_properties(self.actuals_dset)[0]
                a_slice = slice(None)
                if a_shape != f_shape:
                    a_slice = self._get_actuals_slice()
                    logger.debug('Extracting {} sub-slice of {} to match {} '
                                 'shape: {}'.format(a_slice, self.actuals_dset,
                                                    self.fcst_dset, f_shape))

        return a_slice

    def correct_dset(self, out_h5, fcst_perc=None):
        """
        Bias correct and blend (if requested) forecasts:
        - Bias correct forecast data using bias correction factor:
        total actual generation / total forecasted generation
        - Blend fcst_perc of forecast generation with (1 - fcst_perc) of
        actuals generation

        Parameters
        ----------
        out_h5 : str
            Output path for corrected .h5 file
        fcst_perc : float, optional
            Percentage of forecast to use for blending, by default None
        """
        if not os.path.exists(out_h5):
            logger.debug('Copying forecasts ({}) to output path ({})'
                         .format(self.fcst_h5, out_h5))
            shutil.copy(self.fcst_h5, out_h5)

        with h5py.File(out_h5, 'a') as f_out:
            with Resource(self.fcst_h5, unscale=False) as f_fcst:
                fcst = f_fcst[self.fcst_dset]

            with Resource(self.actuals_h5, unscale=False) as f_act:
                actuals = f_act[self.actuals_dset, self._a_slice]

            logger.info('Correcting {} forecasts'.format(self.fcst_dset))
            f_out[self.fcst_dset][...] = self._correct(fcst, actuals,
                                                       fcst_perc=fcst_perc)

    @classmethod
    def correct(cls, fcst_h5, fcst_dset, out_h5,
                actuals_h5=None, actuals_dset=None, fcst_perc=None):
        """
        Bias correct and blend (if requested) forecasts using actuals:
        - Bias correct forecast data using bias correction factor:
        total actual generation / total forecasted generation
        - Blend fcst_perc of forecast generation with (1 - fcst_perc) of
        actuals generation

        Parameters
        ----------
        fcst_h5 : str
            Path to forecast .h5 file
        fcst_dset : str
            Dataset to correct
        out_h5 : str
            Output path for corrected .h5 file
        actuals_h5 : str, optional
            Path to forecast to .h5 file, by default None
        actuals_dset : str, optional
            Actuals dataset, by default None
        fcst_perc : float, None
            Percentage of forecast to use for blending, by default None
        """
        fcst = cls(fcst_h5, fcst_dset, actuals_h5=actuals_h5,
                   actuals_dset=actuals_dset)
        fcst.correct_dset(out_h5, fcst_perc=fcst_perc)

    @classmethod
    def bias_correct(cls, fcst_h5, fcst_dset, out_h5,
                     actuals_h5=None, actuals_dset=None):
        """
        Bias correct forecast using actuals using bias correction factor:
        total actual generation / total forecasted generation

        Parameters
        ----------
        fcst_h5 : str
            Path to forecast .h5 file
        fcst_dset : str
            Dataset to correct
        out_h5 : str
            Output path for corrected .h5 file
        actuals_h5 : str, optional
            Path to forecast to .h5 file, by default None
        actuals_dset : str, optional
            Actuals dataset, by default None
        """
        cls.correct(fcst_h5, fcst_dset, out_h5,
                    actuals_h5=actuals_h5, actuals_dset=actuals_dset)

    @classmethod
    def blend(cls, fcst_h5, fcst_dset, out_h5, fcst_perc,
              actuals_h5=None, actuals_dset=None):
        """
        Bias correct and blend forecast using actuals:
        - Bias correct forecast data using bias correction factor:
        total actual generation / total forecasted generation
        - Blend fcst_perc of forecast generation with (1 - fcst_perc) of
        actuals generation

        Parameters
        ----------
        fcst_h5 : str
            Path to forecast .h5 file
        fcst_dset : str
            Dataset to correct
        out_h5 : str
            Output path for corrected .h5 file
        fcst_perc : float
            Percentage of forecast to use for blending
        actuals_h5 : str, optional
            Path to forecast to .h5 file, by default None
        actuals_dset : str, optional
            Actuals dataset, by default None
        """
        cls.correct(fcst_h5, fcst_dset, out_h5,
                    actuals_h5=actuals_h5, actuals_dset=actuals_dset,
                    fcst_perc=fcst_perc)



# -*- coding: utf-8 -*-
"""
Handler to convert exclusion to/from .h5 and .geotiff
"""
from abc import ABC, abstractstaticmethod
from concurrent.futures import as_completed
import fiona
import geopandas as gpd
import h5py
import json
import logging
import numpy as np
import os
from rasterio import features
import re
from shapely.geometry import shape
from warnings import warn

from rex.utilities import parse_table, SpawnProcessPool
from reV.handlers.exclusions import ExclusionLayers

logger = logging.getLogger(__name__)


class BaseWindSetbacks(ABC):
    """
    Create exclusions layers for wind setbacks
    """
    MULTIPLIERS = {'high': 3, 'moderate': 1.1}

    def __init__(self, excl_h5, hub_height, rotor_diameter, regs_fpath=None,
                 multiplier=None, hsds=False, chunks=(128, 128)):
        """
        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        """
        self._excl_h5 = excl_h5
        self._hub_height = hub_height
        self._rotor_diameter = rotor_diameter
        self._hsds = hsds
        self._shape, self._chunks, self._profile = \
            self._parse_excl_properties(excl_h5, chunks, hsds=hsds)

        self._regs, self._multi = self._preflight_check(regs_fpath, multiplier)

    def __repr__(self):
        msg = "{} for {}".format(self.__class__.__name__, self._excl_h5)
        return msg

    @property
    def hub_height(self):
        """
        Turbine hub-height in meters

        Returns
        -------
        float
        """
        return self._hub_height

    @property
    def rotor_diameter(self):
        """
        Turbine rotor diameter in meters

        Returns
        -------
        float
        """
        return self._rotor_diameter

    @property
    def tip_height(self):
        """
        Turbine blade tip height in meters

        Returns
        -------
        float
        """
        return self._hub_height + self._rotor_diameter / 2

    @property
    def generic_setback(self):
        """
        Default setback of turbine tip height * multiplier, used for global
        setbacks

        Returns
        -------
        float
        """
        if self._multi is None:
            setback = None
        else:
            setback = self.tip_height * self._multi

        return setback

    @staticmethod
    def _parse_excl_properties(excl_h5, chunks, hsds=False):
        """
        Parse exclusions shape, chunk size, and profile from excl_h5 file

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        chunks : tuple | None
            Chunk size of exclusions datasets
        hsds : bool, optional
            Boolean flag to use h5pyd to handle .h5 'files' hosted on AWS
            behind HSDS, by default False

        Returns
        -------
        shape : tuple
            Shape of exclusions datasets
        chunks : tuple | None
            Chunk size of exclusions datasets
        profile : str
            GeoTiff profile for exclusions datasets
        """
        with ExclusionLayers(excl_h5, hsds=hsds) as exc:
            shape = exc.shape
            profile = exc.profile
            if chunks is None:
                chunks = exc.chunks

        if len(chunks) < 3:
            chunks = (1, ) + chunks

        if len(shape) < 3:
            shape = (1, ) + shape

        logger.debug('Exclusions properties:\n'
                     'shape : {}\n'
                     'chunks : {}\n'
                     'profile : {}\n'
                     .format(shape, chunks, profile))

        return shape, chunks, profile

    @abstractstaticmethod
    def _parse_regs(regs_fpath):
        """
        Parse wind regulations

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = parse_table(regs_fpath)
        if 'FIPS' not in regs:
            msg = ('Wind regulations does not have county FIPS! Please add a '
                   '"FIPS" columns with the unique county FIPS values.')
            logger.error(msg)
            raise RuntimeError(msg)

        return regs

    @classmethod
    def _parse_county_regs(cls, regs_fpath, excl_h5):
        """
        Parse wind regulations, combine with county geometries from
        exclusions .h5 file. The county geometries are intersected with
        features to compute county specific setbacks.

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file, must have a county FIPS column
            to match with "cnty_fips" exclusion layer.
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers, one layer must be 'cnty_fips'

        Returns
        -------
        regs: geopandas.GeoDataFrame
            GeoDataFrame with county level wind setback regulations merged
            with county geometries, use for intersecting with setback features
        """
        regs = cls._parse_regs(regs_fpath)

        with ExclusionLayers(excl_h5) as exc:
            fips = exc['cnty_fips']
            transform = exc.profile['transform']
            crs = exc.crs

        fips_df = gpd.GeoDataFrame(columns=['geometry', 'FIPS'], crs=crs)
        for i, (p, v) in enumerate(features.shapes(fips, transform=transform)):
            fips_df.at[i] = shape(p), v

        fips_check = regs['FIPS'].isin(fips_df['FIPS'])
        if not fips_check.all():
            msg = ('The following county FIPS were requested in by the wind '
                   'regulations but were not availble in the Exclusions '
                   '"cnty_fips" layer:\n{}'
                   .format(regs.loc[~fips_check, 'FIPS']))
            logger.error(msg)
            raise RuntimeError(msg)

        regs = fips_df.merge(regs, on='FIPS', how='right')
        logger.debug('Wind regulations were provided for {} counties'
                     .format(len(regs)))

        return regs

    @staticmethod
    def _get_setback(cnty_regs, hub_height, rotor_diameter):
        """
        Compute the setback distance in meters from the county regulations,
        turbine tip height or rotor diameter

        Parameters
        ----------
        cnty_regs : pandas.Series
            Pandas Series with wind regulations for a single county / feature
            type
        hub_height : int | float
            Turbine hub-height, used to compute setbacks from hub-height
            multiplier regulations and to compute tip-height for tip-height
            multiplier regulations
        rotor_diameter : int | float
            Turbine rotor diameter, used to compute setbacks from
            rotor-diameter multiplier regulations

        Returns
        -------
        setback : float | None
            setback distance in meters, None if the setback "Value Type"
            was not recognized
        """
        tip_height = hub_height + rotor_diameter / 2

        setback_type = cnty_regs['Value Type']
        setback = cnty_regs['Value']
        if setback_type == 'Max-tip Height Multiplier':
            setback *= tip_height
        elif setback_type == 'Rotor-Diameter Multiplier':
            setback *= rotor_diameter
        elif setback_type == 'Hub-height Multiplier':
            setback *= hub_height
        elif setback_type != 'Meters':
            msg = ('Cannot create setback for {}, expecting '
                   '"Max-tip Height Multiplier", '
                   '"Rotor-Diameter Multiplier", '
                   '"Hub-height Multiplier", or '
                   '"Meters", but got {}'
                   .format(cnty_regs['County'], setback_type))
            logger.warning(msg)
            warn(msg)
            setback = None

        return setback

    @abstractstaticmethod
    def _parse_features(features_fpath, crs):
        """
        Abstract method to parse features

        Parameters
        ----------
        features_fpath : str
            Path to file containing features to setback from
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        features : geopandas.GeoDataFrame
            Geometries of features to setback from in exclusion coordinate
            system
        """

    @classmethod
    def _map_features_dir(cls, features_dir):
        """
        Map features files to state based on file name

        Parameters
        ----------
        features_dir : str
            Path to directory containing features files. Used
            to identify features to build setbacks from. Files should be
            by state

        Returns
        -------
        features_state_map : dict
            Dictionary mapping state to features file path
        """

    @classmethod
    def _compute_generic_setbacks(cls, features_fpath, crs, setback):
        """
        Abstract method to compute generic setbacks

        Parameters
        ----------
        features_fpath : str
            Path to file containing features to setback from
        crs : str
            Coordinate reference system to convert structures geometries into
        setback : float
            Generic set back distance in meters

        Returns
        -------
        setbacks : list
            List of setback geometries for given features
        """
        features = cls._parse_features(features_fpath, crs)

        features['geometry'] = features.buffer(setback)
        setbacks = [(geom, 1) for geom in features['geometry']]

        return setbacks

    @classmethod
    def _compute_local_setbacks(cls, features_fpath, crs, wind_regs,
                                hub_height, rotor_diameter):
        """
        Compute local features setbacks

        Parameters
        ----------
        features_fpath : str
            Path to file containg feature to setback from
        crs : str
            Coordinate reference system to convert structures geometries into
        wind_regs : pandas.DataFrame
            Wind regulations that define setbacks by county
        hub_height : float
            Turbine hub height in meters
        rotor_diameter : float
            Turbine rotor diameter in meters

        Returns
        -------
        setbacks : list
            List of setback geometries
        """
        features = cls._parse_features(features_fpath, crs)

        setbacks = []
        for i in range(len(wind_regs)):
            cnty = wind_regs.iloc[[i]]
            setback = cls._get_setback(cnty.iloc[0], hub_height,
                                       rotor_diameter)
            if setback is not None:
                logger.debug('- Computing setbacks for county FIPS {}'
                             .format(cnty.iloc[0]['FIPS']))
                tmp = gpd.sjoin(features, cnty,
                                how='inner', op='intersects')
                tmp['geometry'] = tmp.buffer(setback)

                setbacks.extend((geom, 1) for geom in tmp['geometry'])

        return setbacks

    def _preflight_check(self, regs_fpath, multiplier):
        """
        Run preflight checks on WindSetBack inputs:
        1) Ensure either a wind regulations .csv is provided, or
           a setback multiplier
        2) Ensure wind regulations has county FIPS, map regulations to county
           geometries from exclusions .h5 file
        3) Ensure multiplier is a valid entry, either a float or one of
           {'high': 3, 'moderate': 1.1}

        Parameters
        ----------
        regs_fpath : str | None
            Path to wind regulations .csv file, if None create global
            setbacks
        multiplier : int | float | str | None
            setback multiplier to use if wind regulations are not supplied,
            if str, must one of {'high': 3, 'moderate': 1.1}

        Returns
        -------
        regs : geopands.GeoDataFrame | None
            Wind Regulations
        Multiplier : float | None
            Generic setbacks multiplier
        """
        if regs_fpath:
            if multiplier:
                msg = ('A wind regulation .csv file was also provided and '
                       'will be used to determine setback multipliers!')
                logger.warning(msg)
                warn(msg)

            multiplier = None
            regs = self._parse_county_regs(regs_fpath, self._excl_h5)
            logger.debug('Computing setbacks using regulations provided in: {}'
                         .format(regs_fpath))
        elif multiplier:

            regs = None
            if isinstance(multiplier, str):
                multiplier = self.MULTIPLIERS[multiplier]

            logger.debug('Computing setbacks using generic Max-tip Height '
                         'Multiplier of {}'.format(multiplier))
        else:
            msg = ('Computing setbacks requires either a wind regulations '
                   '.csv file or a generic multiplier!')
            logger.error(msg)
            raise RuntimeError(msg)

        return regs, multiplier

    def _rasterize_setbacks(self, shapes):
        """
        Convert setbacks geometries into exclusions array

        Parameters
        ----------
        setbacks : list
            List of (geometry, 1) pairs to rasterize. Each geometry is a
            feature buffered by the desired setback distance in meters

        Returns
        -------
        arr : ndarray
            Exclusions layer arr with proper shape to write to self._excl_h5
        """
        logger.debug('Generating setbacks exclusion array of shape {}'
                     .format(self._shape))
        arr = features.rasterize(shapes=shapes,
                                 out_shape=self._shape[1:],
                                 fill=0,
                                 transform=self._profile['transform'],
                                 dtype='uint8')

        return np.expand_dims(arr, axis=0)

    def _generic_setbacks(self, features, max_workers=None):
        """
        Compute generic setbacks for all states either in serial or parallel

        Parameters
        ----------
        features : list
            list of feature files to compute setbacks from
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks, ready to be written to the exclusions
            .h5 file as a new exclusion layer
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        crs = self._profile['crs']

        setbacks = []
        if max_workers > 1:
            logger.info('Computing generic setbacks in parallel using {} '
                        'workers'.format(max_workers))
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for state_features in features:
                    future = exe.submit(self._compute_generic_setbacks,
                                        state_features, crs,
                                        self.generic_setback)
                    futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    setbacks.extend(future.result())
                    logger.debug('Computed setbacks for {} of {} states'
                                 .format((i + 1), len(features)))
        else:
            logger.info('Computing generic setbacks in serial')
            for i, state_features in enumerate(features):
                setbacks.extend(self._compute_generic_setbacks(
                    state_features, crs, self.generic_setback))
                logger.debug('Computed setbacks for {} of {} states'
                             .format((i + 1), len(features)))

        setbacks = self._rasterize_setbacks(setbacks)

        return setbacks

    def _local_setbacks(self, features_state_map, groupby,
                        max_workers=None):
        """
        Compute local setbacks for all states either in serial or parallel.
        Setbacks are based on local regulation supplied in the wind regulations
        .csv file

        Parameters
        ----------
        features_state_map : dict
            Dictionary mapping features files to states
        groupby : str
            Column to groupby regulations on and map to features files,
            for structures typically "State", for roads typically "Abbr"
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks, ready to be written to the exclusions
            .h5 file as a new exclusion layer
        """
        if max_workers is None:
            max_workers = os.cpu_count()

        regs = self._regs.groupby(groupby)
        crs = self._profile['crs']

        setbacks = []
        if max_workers > 1:
            logger.info('Computing local setbacks in parallel using {} '
                        'workers'.format(max_workers))
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                for state, state_regs in regs:
                    if state in features_state_map:
                        features_fpath = features_state_map[state]
                        future = exe.submit(self._compute_local_setbacks,
                                            features_fpath, crs, state_regs,
                                            self.hub_height,
                                            self.rotor_diameter)
                        futures.append(future)

                for i, future in enumerate(as_completed(futures)):
                    setbacks.extend(future.result())
                    logger.debug('Computed setbacks for {} of {} states'
                                 .format((i + 1), len(futures)))
        else:
            logger.info('Computing local setbacks in serial')
            for i, (state, state_regs) in enumerate(regs):
                if state in features_state_map:
                    features_fpath = features_state_map[state]
                    setbacks.extend(self._compute_local_setbacks(
                        features_fpath, crs, state_regs,
                        self.hub_height, self.rotor_diameter))
                    logger.debug('Computed setbacks for {} of {} states'
                                 .format((i + 1), len(regs)))

        setbacks = self._rasterize_setbacks(setbacks)

        return setbacks

    def _write_layer(self, layer, arr, description=None, replace=False):
        """
        Write exclusion layer to exclusions .h5 file

        Parameters
        ----------
        layer : str
            Exclusion layer name (dataset name)
        arr : ndarray
            Exclusion layers array
        description : str, optional
            Description of exclusion layer (set as an attribute),
            by default None
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False
        """
        if self._hsds:
            msg = ('Cannot write new layers to an exclusion file hosted in '
                   'the cloud behind HSDS!')
            logger.error(msg)
            raise RuntimeError(msg)

        with h5py.File(self._excl_h5, mode='a') as f:
            if layer in f:
                msg = "{} is already present in {}".format(layer,
                                                           self._excl_h5)
                if replace:
                    msg += ', layer data will be replaced with new setbacks'
                    logger.warning(msg)
                    warn(msg)

                    f[layer][...] = arr
                else:
                    msg += ' to replace layer data, set replace=True'
                    logger.error(msg)
                    raise RuntimeError(msg)
            else:
                ds = f.create_dataset(layer,
                                      shape=arr.shape,
                                      dtype=arr.dtype,
                                      chunks=self._chunks,
                                      data=arr)
                logger.debug('\t- {} created and loaded'.format(layer))
                ds.attrs['profile'] = json.dumps(self._profile)
                if description is not None:
                    ds.attrs['description'] = description
                    logger.debug('\t- Description for {} added:\n{}'
                                 .format(layer, description))

    def compute_setbacks(self, features_dir, groupby, max_workers=None,
                         layer=None, description=None, replace=False):
        """
        Compute setbacks for all states either in serial or parallel.
        Existing setbacks are computed if a wind regulations file was supplied
        during class initialization, otherwise generic setbacks are computed

        Parameters
        ----------
        features_dir : str
            Path to directory containing features files. Used
            to identify features to build setbacks from. Files should be
            by state
        groupby : str
            Column to groupby regulations on and map to features files,
            for structures typically "State", for roads typically "Abbr"
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        layer : str, optional
            Name of new layer to write to exclusions .h5 file containing
            computed setbacks, if None do not write to disc, by default None
        description : str, optional
            Description of exclusion layer (set as an attribute),
            by default None
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks, ready to be written to the exclusions
            .h5 file as a new exclusion layer
        """
        features_state_map = self._map_features_dir(features_dir)
        if self._regs is not None:
            setbacks = self._local_setbacks(features_state_map, groupby,
                                            max_workers=max_workers)
        else:
            setbacks = self._generic_setbacks(features_state_map.values(),
                                              max_workers=max_workers)

        if layer is not None:
            logger.debug('Writing setbacks to {} as layer {}'
                         .format(self._excl_h5, layer))
            self._write_layer(layer, setbacks, description=description,
                              replace=replace)

        return setbacks


class StructureWindSetbacks(BaseWindSetbacks):
    """
    Structure Wind setbacks
    """
    @staticmethod
    def _parse_regs(regs_fpath):
        """
        Parse wind regulations, reduce table to just structures

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = BaseWindSetbacks._parse_regs(regs_fpath)

        mask = ((regs['Feature Type'] == 'Structures')
                & (regs['Comment'] != 'Occupied Community Buildings'))
        regs = regs.loc[mask]

        return regs

    @staticmethod
    def _split_state_name(state_name):
        """
        Split state name at capitals to map .geojson files to regulations
        state names

        Parameters
        ----------
        state_name : str
            State name from geojson files paths with out spaces

        Returns
        -------
        str
            State names with spaces added between Capitals (names) to match
            wind regulations state names
        """
        state_name = ' '.join(a for a
                              in re.split(r'([A-Z][a-z]*)', state_name)
                              if a)

        return state_name

    @classmethod
    def _map_features_dir(cls, structures_dir):
        """
        Map structures .geojson files to state based on file name

        Parameters
        ----------
        structure_dir : str
            Path to directory containing microsoft strucutes *.geojsons. Used
            to identify structures to build setbacks from. Files should be
            by state

        Returns
        -------
        structure_state_map : dict
            Dictionary mapping state to geojson file path
        """
        features_state_map = {}
        for file in os.listdir(structures_dir):
            if file.endswith('.geojson'):
                state = file.split('.')[0]
                state = cls._split_state_name(state)
                features_state_map[state] = os.path.join(structures_dir, file)

        return features_state_map

    @staticmethod
    def _parse_features(structure_fpath, crs):
        """
        Load structures from geojson, convert to exclusions coordinate system

        Parameters
        ----------
        structure_fpath : str
            Path to Microsoft .geojson of structures in a given state
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        structures : geopandas.GeoDataFrame
            Geometries for structures in geojson, in exclusion coordinate
            system
        """
        structures = gpd.read_file(structure_fpath, crs=crs)

        return structures.to_crs(crs=crs)

    @classmethod
    def run(cls, excl_h5, structures_dir, layer_name, hub_height,
            rotor_diameter, regs_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, description=None,
            replace=False):
        """
        Compute structural setbacks and write them as a new layer to the
        exclusions .h5 file. If a wind regulations file is given compute
        local setbacks, otherwise compute generic setbacks using the given
        multiplier and the turbine tip-height. File must be locally on disc to
        allow for writing of new layer.

        Parametersß
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        structure_dir : str
            Path to directory containing microsoft strucutes *.geojsons. Used
            to identify structures to build setbacks from. Files should be
            by state
        layer_name : str
            Name of new layer to write to exclusions .h5 file containing
            computed setbacks
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        description : str, optional
            Description of exclusion layer (set as an attribute),
            by default None
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False
        """
        setbacks = cls(excl_h5, hub_height, rotor_diameter,
                       regs_fpath=regs_fpath, multiplier=multiplier,
                       hsds=False, chunks=chunks)
        setbacks.compute_setbacks(structures_dir, "State", layer=layer_name,
                                  max_workers=max_workers,
                                  description=description,
                                  replace=replace)


class RoadWindSetbacks(BaseWindSetbacks):
    """
    Road Wind setbacks
    """
    @staticmethod
    def _parse_regs(regs_fpath):
        """
        Parse wind regulations, reduce table to just roads

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = BaseWindSetbacks._parse_regs(regs_fpath)

        mask = regs['Feature Type'].isin(['Roads', 'Highways', 'Highways 111'])
        regs = regs.loc[mask]

        return regs

    @classmethod
    def _map_features_dir(cls, roads_dir):
        """
        Map roads .gdb files to state based on file name

        Parameters
        ----------
        roads_dir : str
            Path to directory containing here streets *.gdb files. Used
            to identify roads to build setbacks from. Files should be
            by state

        Returns
        -------
        roads_state_map : dict
            Dictionary mapping state to gdb file path
        """
        roads_state_map = {}
        for file in os.listdir(roads_dir):
            if file.endswith('.gdb') and file.startswith('Streets_USA'):
                state = file.split('.')[0].split('_')[-1]
                roads_state_map[state] = os.path.join(roads_dir, file)

        return roads_state_map

    @staticmethod
    def _parse_features(roads_fpath, crs):
        """
        Load roads from gdb file, convert to exclusions coordinate system

        Parameters
        ----------
        roads_fpath : str
            Path to here streets gdb file for given state
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        roads : geopandas.GeoDataFrame.sindex
            Geometries for roads in gdb file, in exclusion coordinate
            system
        """
        lyr = fiona.listlayers(roads_fpath)[0]
        roads = gpd.read_file(roads_fpath, driver='FileGDB', layer=lyr,
                              crs=crs)

        return roads.to_crs(crs=crs)

    @classmethod
    def _compute_local_setbacks(cls, roads_fpath, crs, wind_regs,
                                hub_height, rotor_diameter):
        """
        Compute local road setbacks

        Parameters
        ----------
        roads_fpath : str
            Path to here streets gdb file with roads to setback from
        crs : str
            Coordinate reference system to convert structures geometries into
        wind_regs : pandas.DataFrame
            Wind regulations by county
        hub_height : float
            Turbine hub height in meters
        rotor_diameter : float
            Turbine rotor diameter in meters

        Returns
        -------
        setbacks : list
            List of setback geometries for given eatures
        """
        features = cls._parse_features(roads_fpath, crs)
        features = features[['StreetName', 'geometry']]
        si = features.sindex

        setbacks = []
        for i in range(len(wind_regs)):
            cnty = wind_regs.iloc[i]
            setback = cls._get_setback(cnty, hub_height,
                                       rotor_diameter)
            if setback is not None:
                logger.debug('- Computing setbacks for county FIPS {}'
                             .format(cnty['FIPS']))
                # spatial index bounding box and intersection
                bb_index = \
                    list(si.intersection(cnty['geometry'].bounds))
                bb_pos = features.iloc[bb_index]
                tmp = bb_pos.intersection(cnty['geometry'])
                tmp = gpd.GeoDataFrame(geometry=tmp)
                tmp = tmp[~tmp.is_empty]
                # Buffer setback
                tmp['geometry'] = tmp.buffer(setback)

                setbacks.extend((geom, 1) for geom in tmp['geometry'])

        return setbacks

    @classmethod
    def run(cls, excl_h5, roads_dir, layer_name, hub_height,
            rotor_diameter, regs_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, description=None,
            replace=False):
        """
        Compute road setbacks and write them as a new layer to the
        exclusions .h5 file. If a wind regulations file is given compute
        local setbacks, otherwise compute generic setbacks using the given
        multiplier and the turbine tip-height. File must be locally on disc to
        allow for writing of new layer.

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        roads_dir : str
            Path to directory containing here streets *.gdb files. Used
            to identify roads to build setbacks from. Files should be
            by state
        layer_name : str
            Name of new layer to write to exclusions .h5 file containing
            computed setbacks
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        description : str, optional
            Description of exclusion layer (set as an attribute),
            by default None
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False
        """
        setbacks = cls(excl_h5, hub_height, rotor_diameter,
                       regs_fpath=regs_fpath, multiplier=multiplier,
                       hsds=False, chunks=chunks)
        setbacks.compute_setbacks(roads_dir, 'Abbr', layer=layer_name,
                                  max_workers=max_workers,
                                  description=description,
                                  replace=replace)


class TransmissionWindSetbacks(BaseWindSetbacks):
    """
    Transmission Wind setbacks, computed against a single set of transmission
    features instead of against state level features
    """
    @staticmethod
    def _parse_regs(regs_fpath):
        """
        Parse wind regulations, reduce table to just transmission

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = BaseWindSetbacks._parse_regs(regs_fpath)

        mask = regs['Feature Type'] == 'Transmission'
        regs = regs.loc[mask]

        return regs

    @staticmethod
    def _parse_features(transmission_fpath, crs):
        """
        Load transmission shape file, convert to exclusions coordinate system

        Parameters
        ----------
        transmission_fpath : str
            Path to transmission shape file
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        trans : geopandas.GeoDataFrame.sindex
            Geometries for transmission features, in exclusion coordinate
            system
        """
        trans = gpd.read_file(transmission_fpath, crs=crs)

        return trans.to_crs(crs=crs)

    @classmethod
    def _compute_local_setbacks(cls, features_fpath, crs, cnty,
                                hub_height, rotor_diameter):
        """
        Compute local county setbacks

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from
        crs : str
            Coordinate reference system to convert structures geometries into
        cnty : geopandas.GeoDataFrame
            Wind regulations for a single county
        hub_height : float
            Turbine hub height in meters
        rotor_diameter : float
            Turbine rotor diameter in meters

        Returns
        -------
        setbacks : list
            List of setback geometries for given state
        """
        features = cls._parse_features(features_fpath, crs)

        setbacks = []
        setback = cls._get_setback(cnty.iloc[0], hub_height, rotor_diameter)
        if setback is not None:
            # clip the transmission lines to county geometry
            # pylint: disable=assignment-from-no-return
            tmp = gpd.clip(features, cnty)
            tmp = tmp[~tmp.is_empty]

            # Buffer setback
            tmp['geometry'] = tmp.buffer(setback)

            setbacks.extend((geom, 1) for geom in tmp['geometry'])

        return setbacks

    def compute_setbacks(self, features_fpath, max_workers=None,
                         layer=None, description=None, replace=False):
        """
        Compute setbacks for all states either in serial or parallel.
        Existing setbacks are computed if a wind regulations file was supplied
        during class initialization, otherwise generic setbacks are computed

        Parameters
        ----------
        features_fpath : str
            Path to shape file with features to compute setbacks from
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        layer : str, optional
            Name of new layer to write to exclusions .h5 file containing
            computed setbacks, if None do not write to disc, by default None
        description : str, optional
            Description of exclusion layer (set as an attribute),
            by default None
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False

        Returns
        -------
        setbacks : ndarray
            Raster array of setbacks, ready to be written to the exclusions
            .h5 file as a new exclusion layer
        """
        crs = self._profile['crs']
        setbacks = []
        if self._regs is not None:
            if max_workers is None:
                max_workers = os.cpu_count()

            if max_workers > 1:
                logger.info('Computing local setbacks in parallel using {} '
                            'workers'.format(max_workers))
                loggers = [__name__, 'reVX']
                with SpawnProcessPool(max_workers=max_workers,
                                      loggers=loggers) as exe:
                    futures = []
                    for i in range(len(self._regs)):
                        cnty = self._regs.iloc[[i]]
                        future = exe.submit(self._compute_local_setbacks,
                                            features_fpath, crs, cnty,
                                            self.hub_height,
                                            self.rotor_diameter)
                        futures.append(future)

                    for i, future in enumerate(as_completed(futures)):
                        setbacks.extend(future.result())
                        logger.debug('Computed setbacks for {} of {} states'
                                     .format((i + 1), len(self._regs)))
            else:
                logger.info('Computing local setbacks in serial')
                for i in range(len(self._regs)):
                    cnty = self._regs.iloc[[i]]
                    setbacks.extend(self._compute_local_setbacks(
                        features_fpath, crs, cnty, self.hub_height,
                        self.rotor_diameter))
                    logger.debug('Computed setbacks for {} of {} states'
                                 .format((i + 1), len(self._regs)))
        else:
            logger.info('Computing generic setbacks')
            setbacks.extend(self._compute_generic_setbacks(
                features_fpath, crs, self.generic_setback))

        setbacks = self._rasterize_setbacks(setbacks)

        if layer is not None:
            logger.debug('Writing setbacks to {} as layer {}'
                         .format(self._excl_h5, layer))
            self._write_layer(layer, setbacks, description=description,
                              replace=replace)

        return setbacks

    @classmethod
    def run(cls, excl_h5, features_fpath, layer_name, hub_height,
            rotor_diameter, regs_fpath=None, multiplier=None,
            chunks=(128, 128), max_workers=None, description=None,
            replace=False):
        """
        Compute setbacks and write them as a new layer to the
        exclusions .h5 file. If a wind regulations file is given compute
        local setbacks, otherwise compute generic setbacks using the given
        multiplier and the turbine tip-height. File must be locally on disc to
        allow for writing of new layer.

        Parameters
        ----------
        excl_h5 : str
            Path to .h5 file containing exclusion layers, will also be the
            location of any new setback layers
        features_fpath : str
            Path to shape file with transmission or rail features to compute
            setbacks from
        layer_name : str
            Name of new layer to write to exclusions .h5 file containing
            computed setbacks
        hub_height : float | int
            Turbine hub height (m), used along with rotor diameter to compute
            blade tip height which is used to determine setback distance
        rotor_diameter : float | int
            Turbine rotor diameter (m), used along with hub height to compute
            blade tip height which is used to determine setback distance
        regs_fpath : str | None, optional
            Path to wind regulations .csv file, if None create generic
            setbacks using max-tip height * "multiplier", by default None
        multiplier : int | float | str | None, optional
            setback multiplier to use if wind regulations are not supplied,
            if str, must a key in {'high': 3, 'moderate': 1.1}, if supplied
            along with regs_fpath, will be ignored, multiplied with
            max-tip height, by default None
        chunks : tuple, optional
            Chunk size to use for setback layers, if None use default chunk
            size in excl_h5, by default (128, 128)
        max_workers : int, optional
            Number of workers to use for setback computation, if 1 run in
            serial, if > 1 run in parallel with that many workers, if None
            run in parallel on all available cores, by default None
        description : str, optional
            Description of exclusion layer (set as an attribute),
            by default None
        replace : bool, optional
            Flag to replace local layer data with arr if layer already
            exists in the exlcusion .h5 file, by default False
        """
        setbacks = cls(excl_h5, hub_height, rotor_diameter,
                       regs_fpath=regs_fpath, multiplier=multiplier,
                       hsds=False, chunks=chunks)
        setbacks.compute_setbacks(features_fpath, layer=layer_name,
                                  max_workers=max_workers,
                                  description=description,
                                  replace=replace)


class RailWindSetbacks(TransmissionWindSetbacks):
    """
    Rail Wind setbacks, computed against a single set of railroad features,
    instead of state level features, uses the same approach as
    TransmissionWindSetbacks
    """
    @staticmethod
    def _parse_regs(regs_fpath):
        """
        Parse wind regulations, reduce table to just rail

        Parameters
        ----------
        regs_fpath : str
            Path to wind regulations .csv file

        Returns
        -------
        regs : pandas.DataFrame
            Wind regulations table
        """
        regs = BaseWindSetbacks._parse_regs(regs_fpath)

        mask = regs['Feature Type'] == 'Railroads'
        regs = regs.loc[mask]

        return regs

    @staticmethod
    def _parse_features(rail_fpath, crs):
        """
        Load rail shape file, convert to exclusions coordinate system

        Parameters
        ----------
        rail_fpath : str
            Path to rail shape file
        crs : str
            Coordinate reference system to convert structures geometries into

        Returns
        -------
        rail : geopandas.GeoDataFrame.sindex
            Geometries for rail features, in exclusion coordinate
            system
        """
        rail = gpd.read_file(rail_fpath, crs=crs)

        return rail.to_crs(crs=crs)



# -*- coding: utf-8 -*-
"""
Module to create wind and solar plants for PLEXOS buses
"""
import json
import logging
import numpy as np
import os
import pandas as pd
from sklearn.metrics.pairwise import haversine_distances
from warnings import warn

from rex.rechunk_h5 import to_records_array
from rex.resource import Resource
from rex.utilities import parse_table, SpawnProcessPool

from reVX.handlers.outputs import Outputs
from reVX.handlers.sc_points import SupplyCurvePoints

logger = logging.getLogger(__name__)


class Plants:
    """
    Base class to handle plants
    """
    def __init__(self, plants):
        self._plants = {}

        if isinstance(plants, (np.ndarray, list, tuple)):
            for i, plant in enumerate(plants):
                self[i] = plant
        elif isinstance(plants, dict):
            self._plants = plants

        self._i = 0

    def __repr__(self):
        msg = "{} with {} plants".format(self.__class__.__name__, len(self))
        return msg

    def __len__(self):
        return len(self._plants)

    def __getitem__(self, plant_id):
        """
        Get the plant build out for the given plant_id

        Parameters
        ----------
        plant_id : int | str
            Unique Id for plant of interest

        Returns
        -------
        list
            List of sc_gids' and thier associated resource gids to build plant
            from
        """
        return self._plants.get(plant_id, [])

    def __setitem__(self, plant_id, plant_build):
        """
        Update plant build

        Parameters
        ----------
        plant_id : int | str
            Unique Id for plant to update
        plant_build : list
            List of sc_gids' and thier associated resource gids to build plant
            from
        """
        self._plants[plant_id] = plant_build

    def __iter__(self):
        return self

    def __next__(self):
        if self._i >= len(self._plants):
            self._i = 0
            raise StopIteration

        plant_id = self.plant_ids[self._i]
        self._i += 1

        return self[plant_id]

    @property
    def plant_ids(self):
        """
        Plant ids

        Returns
        -------
        list
        """
        return list(self._plants.keys())

    @property
    def plants(self):
        """
        Dictionary matching plants to plant ids

        Returns
        -------
        dict
        """
        return self._plants

    @property
    def plant_builds(self):
        """
        List of plant builds

        Returns
        -------
        dict
        """
        plant_builds = {pid: pd.concat(plant, axis=1).T
                        for pid, plant in self.plants.items()
                        if plant is not None}

        return plant_builds

    @staticmethod
    def _parse_lists(column):
        """
        Check to see if list values are strings, if so parse with json.loads

        Parameters
        ----------
        column : pandas.Series
            Pandas DataFrame column to check

        Returns
        -------
        column : pandas.Series
            Pandas DataFrame column with values converted to lists if needed
        """
        if isinstance(column.iloc[0], str):
            column = column.apply(json.loads).values

        return column

    @classmethod
    def load(cls, plants_fpath):
        """
        Load pre-filled plants from disc

        Parameters
        ----------
        plants_fpath : str | DataFrame
            DataFrame or path to .csv containing pre-filled plants

        Returns
        -------
        Plants
            Initialized Plants instance with pre-filled plants
        """
        plant_builds = parse_table(plants_fpath)
        if 'plant_id' in plant_builds:
            plant_builds = plant_builds.set_index('plant_id')

        plant_builds = plant_builds.apply(cls._parse_lists)

        plants = {}
        for pid, build in plant_builds.iterrows():
            plant = []
            for i in range(len(build['sc_gids'])):
                sc_point = \
                    pd.Series({'sc_gid': build['sc_gids'][i],
                               'res_gids': build['res_gids'][i],
                               'gid_counts': build['gid_counts'][i],
                               'cf_means': build['res_cf_means'][i],
                               'build_capacity': build['build_capacity'][i]})
                plant.append(sc_point)

            plants[pid] = plant

        return cls(plants)


class PlexosPlants(Plants):
    """
    Class to identify and fill Plants
    """

    def __init__(self, plexos_table, sc_table, cf_fpath, dist_percentile=90,
                 lcoe_col='total_lcoe', lcoe_thresh=1.3, offshore=False,
                 max_workers=None, plants_per_worker=40,
                 points_per_worker=400):
        """
        Parameters
        ----------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        cf_fpath : str
            Path to reV generation output .h5 file
        offshore : bool, optional
            Include offshore points, by default False
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        offshore : bool, optional
            Include offshore points, by default False
        max_workers : int, optional
            Number of workers to use for plant identification, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        """
        self._plant_table = self._parse_plant_table(plexos_table)
        self._capacity = self.plant_table['plant_capacity'].values

        if max_workers is None:
            max_workers = os.cpu_count()

        self._sc_points = \
            SupplyCurvePoints(sc_table, cf_fpath,
                              max_workers=max_workers,
                              points_per_worker=points_per_worker,
                              offshore=offshore)

        plants = self._identify_plants(self.plant_table,
                                       self._sc_points.sc_table,
                                       dist_percentile=dist_percentile,
                                       lcoe_col=lcoe_col,
                                       lcoe_thresh=lcoe_thresh,
                                       max_workers=max_workers,
                                       plants_per_worker=plants_per_worker)

        self._plants = {}
        self._fill_plants(plants)

        self._i = 0

    @property
    def plant_table(self):
        """
        Plants meta data table

        Returns
        -------
        pandas.DataFrame
        """
        return self._plant_table

    @property
    def plant_capacity(self):
        """
        Plant capacities

        Returns
        -------
        ndarray
        """
        return self._capacity

    @property
    def sc_points(self):
        """
        Supply Curve Points

        Returns
        -------
        SupplyCurvePoints
        """
        return self._sc_points

    @staticmethod
    def _parse_plant_table(plexos_table):
        """
        Create Table of unique PLEXOS plants from plexos table

        Parameters
        ----------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table

        Returns
        -------
        plant_table : pandas.DataFrame
        Table of unique plants from plexos table
        """
        plexos_table = parse_table(plexos_table)
        if 'plant_id' not in plexos_table:
            plexos_table = \
                PlantProfileAggregation._parse_plexos_table(plexos_table)

        plant_table = \
            plexos_table.drop_duplicates('plant_id').sort_values('plant_id')

        cols = ['latitude', 'longitude', 'plant_id', 'plant_capacity']
        plant_table = plant_table[cols].set_index('plant_id')

        return plant_table

    @staticmethod
    def _check_coords(coords):
        """
        Check coordinate dimensions and units

        Parameters
        ----------
        coords : ndarray
            Either a single set or an array of (lat, lon) coordinates

        Returns
        -------
        coords : ndarray
            Coordinates in radians
        """
        if len(coords.shape) == 1:
            coords = np.expand_dims(coords, axis=0)

        if np.max(coords) > np.pi or np.min(coords) < - np.pi:
            coords = np.radians(coords)

        return coords

    @classmethod
    def _haversine_dist(cls, plant_coords, sc_coords):
        """
        Compute the haversine distance between the given plant(s) and given
        supply curve points

        Parameters
        ----------
        plant_coords : ndarray
            (lat, lon) coordinates of plant(s)
        sc_coords : ndarray
            n x 2 array of supply curve (lat, lon) coordinates

        Returns
        -------
        dist : ndarray
            Vector of distances between plant and supply curve points in km
        """
        plant_coords = cls._check_coords(plant_coords)
        sc_coords = cls._check_coords(sc_coords)

        dist = haversine_distances(plant_coords, sc_coords)
        if plant_coords.shape[0] == 1:
            dist = dist.flatten()

        R = 6373.0  # radius of the earth in kilometers

        return dist * R

    @staticmethod
    def _substation_distance(sc_table, percentile=90):
        """
        Determine the nth percentile of distance between substations and
        transmission from supply curve table

        Parameters
        ----------
        sc_table : pandas.DataFrame
            Supply curve table
        percentile : int, optional
            Percentile to compute substation to transmission distance for,
            by default 90

        Returns
        -------
        dist
            Nth percentile of distance between substations and transmission in
            km, used as plant search distance threshold
        """
        substations = sc_table['trans_type'] == "Substation"
        dist = sc_table.loc[substations, 'dist_mi'].values * 1.6

        return np.percentile(dist, percentile)

    @classmethod
    def _get_plant_sc_dists(cls, bus_coords, sc_table, dist_percentile=90,
                            lcoe_col='total_lcoe', lcoe_thresh=1.3):
        """
        Extract Supply curve gids and distances for plant originating at
        PLEXOS bus coords

        Parameters
        ----------
        bus_coords : ndarray
            bus (lat, lon) coordinates
        sc_table : pandas.DataFrame
            Supply Curve Table
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3

        Returns
        -------
        plant_sc : pandas.DataFrame
            Supply Curve for plant with distance to each sc_gid appended
        """
        logger.debug("Extracting supply curve gids for bus at {}"
                     .format(bus_coords))
        sc_coords = np.radians(sc_table[['latitude', 'longitude']].values)

        # Filter SC table to points within 'dist_tresh' of coords
        dist = cls._haversine_dist(bus_coords, sc_coords)
        dist_thresh = cls._substation_distance(sc_table,
                                               percentile=dist_percentile)
        logger.debug("- Using distance threshold of {} km".format(dist_thresh))
        while True:
            mask = dist <= dist_thresh
            plant_sc = sc_table[['latitude', 'longitude', lcoe_col]].copy()
            plant_sc = plant_sc.loc[mask]
            if len(plant_sc) > 1:
                break
            else:
                dist_thresh *= 1.2

        # Find lowest lcoe site
        pos = np.argmin(plant_sc[lcoe_col])
        lcoe_thresh = plant_sc.iloc[pos][lcoe_col] * lcoe_thresh
        plant_coords = \
            plant_sc.iloc[pos][['latitude', 'longitude']].values.astype(float)
        logger.debug("- Plant will be centered at {}".format(plant_coords))
        logger.debug("- Only supply curve points with an lcoe < {} will be "
                     "used".format(lcoe_thresh))

        # Filter SC table to lcoe values within 'lcoe_thresh' of min LCOE value
        sc_cols = ['sc_gid', lcoe_col]
        plant_sc = sc_table[sc_cols].copy()
        plant_sc["bus_dist"] = dist
        mask = plant_sc[lcoe_col] <= lcoe_thresh
        plant_sc = plant_sc.loc[mask]
        sc_coords = sc_coords[mask]

        # Sort by distance
        plant_sc['dist'] = cls._haversine_dist(plant_coords, sc_coords)
        plant_sc = plant_sc.sort_values('dist')

        return plant_sc.reset_index(drop=True)

    @classmethod
    def _identify_plants(cls, plant_table, sc_table, dist_percentile=90,
                         lcoe_col='total_lcoe', lcoe_thresh=1.3,
                         max_workers=None, plants_per_worker=40):
        """
        Identify plant associated with each bus and return supply curve table

        Parameters
        ----------
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        max_workers : int, optional
            Number of workers to use for plant identification, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40

        Returns
        -------
        plants : list
            List of supply curve points that can be used to fill each plant of
            interest
        """
        plants = []
        if max_workers is None:
            max_workers = os.cpu_count()

        if max_workers > 1:
            logger.info('Identifying plants in parallel')
            loggers = [__name__, 'reVX']
            with SpawnProcessPool(max_workers=max_workers,
                                  loggers=loggers) as exe:
                futures = []
                slices = SupplyCurvePoints._create_worker_slices(
                    plant_table, points_per_worker=plants_per_worker)
                for table_slice in slices:
                    future = exe.submit(cls._identify_plants,
                                        plant_table.iloc[table_slice].copy(),
                                        sc_table,
                                        dist_percentile=dist_percentile,
                                        lcoe_col=lcoe_col,
                                        lcoe_thresh=lcoe_thresh,
                                        max_workers=1)
                    futures.append(future)

                for i, future in enumerate(futures):
                    plants.extend(future.result())
                    logger.debug('Identified {} out of {} plants'
                                 .format((i + 1) * plants_per_worker,
                                         len(plant_table)))
        else:
            logger.debug('Identifying plants in serial')
            for i, bus in plant_table.iterrows():
                coords = \
                    bus[['latitude', 'longitude']].values.astype(float)
                plant = cls._get_plant_sc_dists(
                    coords, sc_table,
                    dist_percentile=dist_percentile,
                    lcoe_col=lcoe_col,
                    lcoe_thresh=lcoe_thresh)
                plants.append(plant)
                logger.debug('Identified {} out of {} PlexosPlants.'
                             .format(i + 1, len(plant_table)))

        return plants

    @staticmethod
    def _get_sc_gids(identified_plants, idx):
        """
        For all plants extract sc_gid, dist (to sc_gid), and bus_dist from
        Supply Curve points

        Parameters
        ----------
        identified_plants : list
            List of identified plants, I.E., Supply curve points available
            to fill each plant along with the distance to the plant center
            and associated bus.
        idx : int
            index to extract from plant sc_tables

        Returns
        -------
        tuple
            (sc_gids, dists, bus_dists)
        """
        sc_gids = []
        bus_dists = []
        dists = []
        for plant in identified_plants:
            sc_point = plant.iloc[idx]
            sc_gids.append(sc_point['sc_gid'])
            bus_dists.append(sc_point['bus_dist'])
            dists.append(sc_point['dist'])

        bus_dists = np.array(bus_dists)
        dists = np.array(dists)

        return sc_gids, dists, bus_dists

    def _allocate_sc_gids(self, sc_gids, dists, bus_dists):
        """
        Allocate capacity from supply curve points to plants

        Parameters
        ----------
        sc_gids : list
            List of supply curve point gids to allocate capacity from
        dists : list
            List of distances from plants to sc_gids
        bus_dists : list
            List of distances from bus associated with plants to sc_gids
        """
        unique_gids, plant_gids = np.unique(sc_gids, return_inverse=True)
        for i, sc_gid in enumerate(unique_gids):
            sc_gid = int(sc_gid)
            if self.sc_points.check_sc_gid(sc_gid):
                plant_ids = np.where(plant_gids == i)[0]
                if len(plant_ids) > 1:
                    sc_dists = dists[plant_ids]
                    if len(sc_dists) != len(np.unique(sc_dists)):
                        idxs = np.argsort(bus_dists[plant_ids])
                    else:
                        idxs = np.argsort(sc_dists)

                    plant_ids = plant_ids[idxs]

                for plant_id in plant_ids:
                    capacity = self.plant_capacity[plant_id]
                    if (capacity > 0) and self.sc_points.check_sc_gid(sc_gid):
                        sc_point, sc_capacity = \
                            self.sc_points.get_capacity(sc_gid, capacity)
                        if sc_capacity:
                            plant = self[plant_id]
                            plant.append(sc_point)

                            self[plant_id] = plant
                            self._capacity[plant_id] -= sc_capacity
                            logger.debug('Allocating {:.1f}MW to plant {} from'
                                         ' sc_gid {}'.format(sc_capacity,
                                                             plant_id,
                                                             sc_gid))
                        else:
                            msg = ('WARNING: sc_gid {} returned 0 capacity!'
                                   .format(sc_gid))
                            logger.warning(msg)
                            warn(msg)

    def _fill_plants(self, identified_plants):
        """
        Fill plants with capacity from supply curve points

        Parameters
        ----------
        identified_plants : list
            List of identified plants, I.E., Supply curve points available
            to fill each plant along with the distance to the plant center
            and associated bus.
        """
        i = 0
        total_cap = np.sum(self.plant_capacity)
        while np.any(self.plant_capacity > 0):
            i_cap = np.sum(self.plant_capacity[self.plant_capacity > 0])
            logger.info('Allocating sc_gids to plants round {}'
                        .format(i))
            sc_gids, dists, bus_dists = self._get_sc_gids(identified_plants, i)
            self._allocate_sc_gids(sc_gids, dists, bus_dists)
            cap = np.sum(self.plant_capacity[self.plant_capacity > 0])
            logger.info('{:.1f} MW allocated in round {}'
                        .format(i_cap - cap, i))
            i += 1
            logger.info('{:.1f} MW allocated out of {:.1f} MW'
                        .format(total_cap - cap, total_cap))
            logger.info('{} of {} plants have been filled'
                        .format(np.sum(self.plant_capacity <= 0), len(self)))

    def dump(self, out_fpath=None):
        """
        Create plants meta data from filled plants DataFrames:
            - Location (lat, lon)
            - final capacity
            - sc_gids
            - res_gids
            - res gid_counts

        Parameters
        ----------
        out_fpath : str, optional
            .csv path to save plant meta data too, by default None

        Returns
        -------
        plants_meta : pandas.DataFrame
            Location (lat, lon), final capacity, and associated sc_gids,
            res_gids, and res gid_counts for all plants
        """
        plants_meta = []
        for pid, plant in self.plant_builds.items():
            plants_meta.append(pd.Series(
                {'sc_gids': plant['sc_gid'].values.tolist(),
                 'res_gids': plant['res_gids'].values.tolist(),
                 'gid_counts': plant['gid_counts'].values.tolist(),
                 'res_cf_means': plant['cf_means'].values.tolist(),
                 'build_capacity': plant['build_capacity'].values.tolist()},
                name=pid))

        plants_meta = pd.concat(plants_meta, axis=1).T
        plants_meta.index.name = 'plant_id'

        if out_fpath:
            plants_meta.to_csv(out_fpath)

        return plants_meta

    @classmethod
    def save(cls, plexos_table, sc_table, cf_fpath, out_fpath,
             dist_percentile=90, lcoe_col='total_lcoe', lcoe_thresh=1.3,
             offshore=False, max_workers=None, plants_per_worker=40,
             points_per_worker=400):
        """
        Identify, fill, and then save plants to disc

        Parameters
        ----------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        cf_fpath : str
            Path to reV generation output .h5 file
        out_fpath : str
            .csv path to save plant meta data too
        offshore : bool, optional
            Include offshore points, by default False
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        offshore : bool, optional
            Include offshore points, by default False
        max_workers : int, optional
            Number of workers to use for plant identification, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        """
        pp = cls(plexos_table, sc_table, cf_fpath,
                 dist_percentile=dist_percentile, lcoe_col=lcoe_col,
                 lcoe_thresh=lcoe_thresh, offshore=offshore,
                 max_workers=max_workers, plants_per_worker=plants_per_worker,
                 points_per_worker=points_per_worker)

        pp.dump(out_fpath=out_fpath)


class PlantProfileAggregation:
    """
    Aggregate renewable generation profiles to Plexos "plants"
    """
    def __init__(self, plexos_table, sc_table, cf_fpath, plants=None,
                 dist_percentile=90, lcoe_col='total_lcoe',
                 lcoe_thresh=1.3, offshore=False, max_workers=None,
                 plants_per_worker=40, points_per_worker=400):
        """
        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity provided as a .csv,
            .json, or pandas DataFrame
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        cf_fpath : str
            Path to reV generation output .h5 file
        offshore : bool, optional
            Include offshore points, by default False
        """
        self._plexos_table = self._parse_plexos_table(plexos_table)
        self._cf_fpath = cf_fpath
        self._cf_gid_map = self._parse_cf_gid_map(cf_fpath)
        self._sc_table = SupplyCurvePoints._parse_sc_table(sc_table,
                                                           offshore=offshore)
        if plants is None:
            self._plants = PlexosPlants(self._plexos_table, self._sc_table,
                                        cf_fpath,
                                        dist_percentile=dist_percentile,
                                        lcoe_col=lcoe_col,
                                        lcoe_thresh=lcoe_thresh,
                                        offshore=offshore,
                                        max_workers=max_workers,
                                        plants_per_worker=plants_per_worker,
                                        points_per_worker=points_per_worker)
        else:
            self._plants = Plants.load(plants)

        self._sc_bus_dist = None

    def __repr__(self):
        msg = "{} with {} plants".format(self.__class__.__name__, len(self))
        return msg

    def __len__(self):
        return len(self.plexos_table)

    @property
    def plexos_table(self):
        """
        PLEXOS table

        Returns
        -------
        pandas.DataFrame
        """
        return self._plexos_table

    @property
    def cf_fpath(self):
        """
        reV generation output file path

        Returns
        -------
        str
        """
        return self._cf_fpath

    @property
    def cf_gid_map(self):
        """
        Mapping of {res_gid: gen_gid}

        Returns
        -------
        dictionary
        """
        return self._cf_gid_map

    @property
    def sc_table(self):
        """
        Supply Curve Table

        Returns
        -------
        pandas.DataFrame
        """
        return self._sc_table

    @property
    def plants(self):
        """
        Dictionary matching plants to plant ids

        Returns
        -------
        dict
        """
        return self._plants.plants

    @property
    def plant_builds(self):
        """
        PLEXOS Plant builds

        Returns
        -------
        dict
        """
        return self._plants.plant_builds

    @property
    def sc_bus_dist(self):
        """
        Compute distance between every Supply Curve gid and every PLEXOS bus

        Returns
        -------
        ndarray
        """
        if self._sc_bus_dist is None:
            cols = ['latitude', 'longitude']
            plant_coords = self._plexos_table[cols].values.astype(float)
            sc_coords = self.sc_table[cols].values.astype(float)

            self._sc_bus_dist = \
                PlexosPlants._haversine_dist(plant_coords, sc_coords).T

        return self._sc_bus_dist

    @staticmethod
    def _parse_plexos_table(plexos_table):
        """
        Parse PLEXOS table from file and reduce to PLEXOS_COLS
        Combine buses at the same coordinates and add unique plant_ids

        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity provided as a .csv,
            .json, or pandas DataFrame

        Returns
        -------
        plexos_table : pandas.DataFrame
            Parsed and clean PLEXOS table
        """
        plexos_table = parse_table(plexos_table)
        cols = ['generator', 'busid', 'busname', 'capacity', 'latitude',
                'longitude', 'system']
        cols = [c for c in plexos_table if c.lower() in cols]
        plexos_table = plexos_table[cols]
        rename = {}
        for c in plexos_table:
            if c.lower() == 'capacity':
                rename[c] = 'capacity'
            elif c.lower() == 'latitude':
                rename[c] = 'latitude'
            elif c.lower() == 'longitude':
                rename[c] = 'longitude'

        plexos_table = plexos_table.rename(columns=rename)

        mask = plexos_table['latitude'] > 90
        mask |= plexos_table['latitude'] < -90
        mask |= plexos_table['longitude'] > 180
        mask |= plexos_table['longitude'] < -180
        if np.any(mask):
            msg = ('WARNING: {} Buses have invalid coordinates:\n{}'
                   .format(np.sum(mask), plexos_table.loc[mask]))
            logger.warning(msg)
            warn(msg)
            plexos_table = plexos_table.loc[~mask]

        mask = plexos_table['capacity'] > 0
        plexos_table = plexos_table.loc[mask]
        cols = ['latitude', 'longitude']
        plant_cap = plexos_table.groupby(cols)['capacity'].sum()
        plant_cap = plant_cap.reset_index().reset_index()
        rename = {'index': 'plant_id', 'capacity': 'plant_capacity'}
        plant_cap = plant_cap.rename(columns=rename)

        plexos_table = plexos_table.merge(plant_cap,
                                          on=cols,
                                          how='inner')

        return plexos_table

    @staticmethod
    def _parse_cf_gid_map(cf_fpath):
        """
        Map resource gids to gen gids

        Parameters
        ----------
        cf_fpath : str
            Path to reV generation output .h5 file

        Returns
        -------
        cf_gid_map : dictionary
            Mapping of {res_gid: gen_gid}
        """
        with Resource(cf_fpath) as f:
            res_gids = f.get_meta_arr('gid')

        if not isinstance(res_gids, np.ndarray):
            res_gids = np.array(list(res_gids))

        cf_gid_map = {gid: i for i, gid in enumerate(res_gids)}

        return cf_gid_map

    def plants_meta(self):
        """
        Create plants meta data from filled plants DataFrames:
            - Location (lat, lon)
            - final capacity
            - sc_gids
            - res_gids
            - res gid_counts
        Parameters
        ----------
        plants : list
            List of filled plant DataFrames
        out_fpath : str
            .csv path to save plant meta data too
        Returns
        -------
        plants_meta : pandas.DataFrame
            Location (lat, lon), final capacity, and associated sc_gids,
            res_gids, and res gid_counts for all plants
        """
        plants_meta = []
        for pid, plant in self.plant_builds.items():
            res_gids = plant['res_gids'].values.tolist()
            plants_meta.append(pd.Series(
                {'sc_gids': plant['sc_gid'].values.tolist(),
                 'res_gids': res_gids,
                 'gid_counts': plant['gid_counts'].values.tolist(),
                 'gen_gids': [[self.cf_gid_map[gid] for gid in gids]
                              for gids in res_gids],
                 'res_cf_means': plant['cf_means'].values.tolist(),
                 'build_capacity': plant['build_capacity'].values.tolist(),
                 'cf_mean': np.hstack(plant['cf_means'].values).mean()},
                name=pid))

        plants_meta = pd.concat(plants_meta, axis=1).T
        plants_meta.index.name = 'plant_id'

        plants_meta = self.plexos_table.merge(plants_meta.reset_index(),
                                              on='plant_id', how='outer')

        return plants_meta

    def _make_plant_meta(self, bus_meta):
        """
        Create plant meta data for given bus

        Parameters
        ----------
        bus_meta : pandas.Series
            Meta data for desired bus to build plant for

        Returns
        -------
        plant_meta : pandas.DataFrame
            Meta data for plant associated with given bus, constructed from:
            - Plant table
            - Supply Curve table
            - Bus capacity
        """
        plant_meta = self.plant_builds[bus_meta['plant_id']]
        plant_meta['gen_gids'] = \
            plant_meta['res_gids'].apply(lambda gids: [self.cf_gid_map[gid]
                                                       for gid in gids])

        sc_cols = ['res_gids', 'gen_gids', 'gid_counts', 'capacity']
        sc_cols = [c for c in self.sc_table if c not in sc_cols]
        plant_meta = plant_meta.merge(self.sc_table[sc_cols],
                                      on='sc_gid', how='left')

        plant_capacity = plant_meta['build_capacity'].sum()
        if plant_capacity != bus_meta['capacity']:
            bulid_capacity = (plant_meta['build_capacity'] / plant_capacity
                              * bus_meta['capacity'])
            plant_meta.loc[:, 'build_capacity'] = bulid_capacity

        return plant_meta

    @staticmethod
    def _make_profile(cf_fpath, plant_build):
        """
        Make generation profiles for given plant buildout

        Parameters
        ----------
        cf_fpath : str
            Path to reV Generation output .h5 file to pull CF profiles from
        plant_build : pandas.DataFrame
            DataFrame describing plant buildout:
                - Supply curve gids and the capacity to build at each
                    - res_gids, gen_gids, gid_counts by sc_gid

        Returns
        -------
        profile: ndarray
            Generation profile for plant as a vector
        """
        with Resource(cf_fpath) as f:
            profile = None
            for _, row in plant_build.iterrows():
                gid_capacities = (row['gid_counts'] / np.sum(row['gid_counts'])
                                  * row['build_capacity'])
                cf_profiles = f['cf_profile', :, row['gen_gids']]
                for i, cf_profile in enumerate(cf_profiles.T):
                    if profile is None:
                        profile = cf_profile * gid_capacities[i]
                    else:
                        profile += cf_profile * gid_capacities[i]

        if len(profile.shape) != 1:
            profile = profile.flatten()

        return profile

    def aggregate_profiles(self, out_fpath):
        """
        Aggregate plants from capacity factor profiles and save to given
        output .h5 path

        Parameters
        ----------
        out_fpath : str
            Output .h5 path to save profiles to
        """
        with Outputs(out_fpath, mode='w') as f_out:
            f_out.set_version_attr()
            with Resource(self.cf_fpath) as f_in:
                logger.info('Copying time_index')
                f_out['time_index'] = f_in.time_index

            logger.info('Writing meta data')
            f_out['meta'] = self.plants_meta()

            f_out.h5.create_group('plant_meta')
            gen_profiles = []
            logger.info('Extracting profiles and writing meta for plants')
            for bus_id, bus_meta in self.plexos_table.iterrows():
                logger.debug('Building plant for bus {}'.format(bus_id))
                plant_meta = self._make_plant_meta(bus_meta)
                gen_profiles.append(self._make_profile(self.cf_fpath,
                                                       plant_meta.copy()))

                plant_meta = to_records_array(plant_meta)
                logger.debug('Writing plant_meta/{}'.format(bus_id))
                f_out._create_dset('plant_meta/{}'.format(bus_id),
                                   plant_meta.shape,
                                   plant_meta.dtype,
                                   chunks=None,
                                   data=plant_meta)

            logger.info('Writing Generation Profiles')
            gen_profiles = \
                np.round(np.dstack(gen_profiles)[0]).astype('uint16')
            f_out._create_dset('gen_profiles',
                               gen_profiles.shape,
                               gen_profiles.dtype,
                               chunks=(None, 100),
                               data=gen_profiles)

    @classmethod
    def aggregate(cls, plexos_table, sc_table, cf_fpath, plants_fpath,
                  out_fpath, offshore=False):
        """
        Aggregate pre-filled plants

        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity provided as a .csv,
            .json, or pandas DataFrame
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        cf_fpath : str
            Path to reV Generation output .h5 file to pull CF profiles from
        plants_fpath : str
            Path to .csv containing pre-filled plants
        out_fpath : str
            .h5 path to save aggregated plant profiles to
        offshore : bool, optional
            Include offshore points, by default False
        """
        pp = cls(plexos_table, sc_table, cf_fpath, plants=plants_fpath,
                 offshore=offshore)
        # Add plants to PlexosPlant instance

        pp.aggregate_profiles(out_fpath)

    @classmethod
    def run(cls, plexos_table, sc_table, cf_fpath, out_fpath,
            dist_percentile=90, lcoe_col='total_lcoe', lcoe_thresh=1.3,
            max_workers=None, points_per_worker=400, plants_per_worker=40,
            offshore=False):
        """
        Find, fill, and save profiles for Plants associated with given PLEXOS
        buses

        Parameters
        ----------
        plexos_table : str | pandas.DataFrame
            PLEXOS table of bus locations and capacity provided as a .csv,
            .json, or pandas DataFrame
        sc_table : str | pandas.DataFrame
            Supply Curve table .csv or pre-loaded pandas DataFrame
        cf_fpath : str
            Path to reV Generation output .h5 file to pull CF profiles from
        out_fpath : str
            .h5 path to save aggregated plant profiles to
        dist_percentile : int, optional
            Percentile to use to compute distance threshold using sc_gid to
            SubStation distance , by default 90
        lcoe_col : str, optional
            LCOE column to sort by, by default 'total_lcoe'
        lcoe_thresh : float, optional
            LCOE threshold multiplier, exclude sc_gids above threshold,
            by default 1.3
        max_workers : int, optional
            Number of workers to use for point and plant creation, 1 == serial,
            > 1 == parallel, None == parallel using all available cpus,
            by default None
        points_per_worker : int, optional
            Number of points to create on each worker, by default 400
        plants_per_worker : int, optional
            Number of plants to identify on each worker, by default 40
        offshore : bool, optional
            Include offshore points, by default False
        """
        pp = cls(plexos_table, sc_table, cf_fpath, offshore=offshore,
                 dist_percentile=dist_percentile, lcoe_col=lcoe_col,
                 lcoe_thresh=lcoe_thresh, max_workers=max_workers,
                 points_per_worker=points_per_worker,
                 plants_per_worker=plants_per_worker)

        pp.aggregate_profiles(out_fpath)
